package binomialHeap.concurrent

import java.util.concurrent.atomic._
import ox.cads.util.{ThreadID,Profiler, Log, NullLog,Backoff}
import scala.collection.mutable.ArrayBuffer
import scala.util.Random

/** A concurrent binary heap. 
  * @param theLog a Log to use for loging.
  * @param MaxRetries the maximum number of retries that loops should attempt.
  * @param numMinsMin the number of minimum values to record while minimum is 
  * traversing.
  * @param numMinsDelMin the number of minimum values to record while deleteMin
  * is traversing.
  * @param tidyRatio the recipricol of the probability with which a thread 
  * should attempt to tidy the heap when it completes an operation. 
  * @param backtracking does deleteMin backtrack if it finds the current node's
  * sequence number has changed? 
  * @param allowUnion is a union operation allowed concurrently with other 
  * operations on the giving heap. */
abstract class ConcHeapBase(
  theLog: Log = NullLog, MaxRetries: Int = 2000,
  numMinsMin: Int = 1, numMinsDelMin: Int = 1,
  tidyRatio: Int = 8, RecTidy: Boolean = true,
  A: Int = 16, B: Int = 16,
  backtracking: Boolean = false, retrieving: Boolean = false,
  allowUnion: Boolean = false)
    extends binomialHeap.PriorityQueue{

  // println(allowUnion)

  // ========= The main state

  /** The head of the list.  Sentinel head node with arbitrary key.  */
  protected[concurrent] val head: Node = new Node(-1)

  val id = scala.util.Random.nextInt(10000)

  /** The epoch, i.e. the number of times a union operation has occurred.  This
    * is incremented only when a UnionGiver label is on head. */
  protected val epoch = new AtomicLong(0)

  /** Is ep equal to the current abstract epoch?
    * Pre: ep was previously obtained using getEpoch (so is not larger than
    *  epoch, and was valid when read). */
  protected def epochValid(ep: Long): Boolean = head.getLabel match{
    case ug: UnionGiver if ug.joined => false
    case _ => epoch.get == ep
  }
 
  /** Optionally get the current abstract epoch.
    * @return None to indicate that the heap is currently empty (because of 
    * a linearized but incomplete union operation); or Some(ep) to indicate
    * that the abstract epoch is ep. */
  protected def getEpoch: Option[Long] = {
    val ep = epoch.get // lin point for this op in successful case.
    head.getLabel match{
      case ug: UnionGiver if ug.joined => None
      case _ => if(epoch.get == ep) Some(ep) else None
    }
  }

  /** An object referencing this heap.  Defined in ConcHeap.scala. */
  protected val heapIndirection: AtomicReference[HeapIndirection]

  // ========= Various helper functions

  /** Log object to use. */
  protected[concurrent] def log(st: => String) = theLog(ThreadID.get+": "+st)

  /** Shut down the heap, in particular the log (if used). */
  def shutdown = theLog.shutdown

  // Random number generators, used in Insert and Tidy.
  private val tlRandom = new ThreadLocal[Random]{
    override def initialValue = new Random(ThreadID.get)
  }
  protected[concurrent] def random = tlRandom.get

  // Exponential back-off objects, used in Insert and DeleteMin.  
  private val tlBackoff = new ThreadLocal[Backoff]{
    override def initialValue = new Backoff(400, 300000, 1.5F, false)
  }
  protected def backoff = tlBackoff.get

  /** Assertions, using the log. */
  protected def myAssert(ok: Boolean, st: => String) : Unit =
    theLog.assert(ok, ThreadID.get+": "+st+"\n"+this)

  protected def myAssert(ok: Boolean) : Unit = myAssert(ok: Boolean, "")

  /** Convert to a string, for logging purposes.  Note this in
    * unlinearizable. */
  override def toString =
    // "STATUS: "+status.get.toString+
    id+": HEAP: "+head.toString(n = 40, d = 1)

  def toString(n: Int) =
    // "STATUS: "+status.get.toString+
    id+": HEAP: "+head.toString(n = n, d = 1)

  /** Approximation to the number of roots. */
  //private val rootCount = new AtomicInteger(0)

  /** Clear the heap.  Assumes mutual exclusion. */
  def clear =  head.clear

  // ======== Adding and removing labels from head nodes.

  /** Try to add label onto head, as long as n.getState = nState and the current
    * epoch is ep.
    * @return true to indicate head successfully labelled; false to indicate 
    * n.getState != nState or the epoch != ep.  */
  protected def labelHead(label: HeapLabel, n: Node, nState: NodeState, ep: Long)
      : Boolean = {
    // log("labelHead on heap "+id+" with label "+label)
    var done = false // has the label been added?
    var ok = true // if ok = false, n.getState != nState or epoch.get != ep
    while(!done && ok){
      if(n.getState != nState || epoch.get != ep) // failed
        ok = false
      else{
        val headState = head.getState
        headState.label match{
          case null => // add new HeapLabelList
            val hll = new HeapLabelList(this, label)
            done = head.getState == headState &&
              head.compareAndSet(headState, headState.addLabel(hll))
            // if(done) log("added new HeapLabelList")
          case hll @ HeapLabelList(h) =>
            assert(h == this)
            done = hll.add(label)
            // if(done) log("Added to existing HeapLabelList "+hll)
            // else log(" failed to add to existing HeapLabelList "+hll)
            if(!done) assert(head.getLabel != hll)
          case l =>
            // log("labelHead helping with "+l)
            help(head, headState)
        } // end of match
      }
    }
    ok
  }

  /** Remove label from head. */
  protected def unlabelHead(label: Label) = {
    val headState = head.getState
    if(headState.label eq label)
      head.compareAndSet(headState, headState.addLabel(null))
    // Otherwise another thread removed the label
    assert(head.getLabel ne label)
  }

  /** Call tidy on the HeadLabelList on this heap head, if any. */
  protected def tidyHeapLabelList = head.getLabel match{
    case hll: HeapLabelList => hll.tidy
    case _ => { } // log("clearInserted didn't tidy")
  }

  // ========= Tidying

  /** Maybe perform tidying.
    * Defined in Tidy.scala. */
  protected def maybeTidy()

  // ========= Helping

  /** Help the label on helpNode, whose state is helpState. 
    * Pre: the label is not null.
    * Defined in ConcHeap. */
  protected[concurrent] def help(helpNode: Node, helpState: NodeState)

  /** Help with a Delete label, if the previous node has appropriate state.
    * @param pred the previous node.
    * @param helpNode the node with the Delete label.
    * @param helpState a recent state of helpNode, with label delLabel. 
    * Defined in DeleteMin. */
  protected def helpDelete(pred: Node, helpNode: Node, helpState: NodeState)

  /** Help with a UnionGiver label. 
    * @param helpNode the node with the Delete label.
    * @param helpState a recent state of helpNode, with label delLabel.
    * @param ug the UnionGiver label. */
  protected def helpUG(helpNode: Node, helpState: NodeState, ug: UnionGiver)

  /** If h is the giving heap in a union with this heap, and the joining CAS has
    * happened, then help to complete that union. */
  // protected def maybeHelpUG(h: ConcHeap)

  /** Help with a HeapLabel. */
  protected[concurrent] def helpHL(label: HeapLabel)

  /** Help to complete a UnionReceiver heap label. */
  protected[concurrent] def helpUnionReceiver(ur: UnionReceiver) 

}
