package binomialHeap.concurrent

import ox.cads.util.{Log,NullLog,ThreadID}

/** Adding the deleteMin operation to the binomial heap. */
abstract class DeleteMin(
  theLog: Log = NullLog, MaxRetries: Int = 2000,
  numMinsMin: Int = 1, numMinsDelMin: Int = 1,
  tidyRatio: Int = 8, RecTidy: Boolean = true,
  A: Int = 16, B: Int = 16,
  backtracking: Boolean = false, retrieving: Boolean = false,
  allowUnion: Boolean = false)
    extends
    Deletion(theLog, MaxRetries, numMinsMin, numMinsDelMin, tidyRatio, 
                 RecTidy, A, B, backtracking, retrieving, allowUnion){

  // ========= deleteMin

  /** Delete the minimum node. */
  def deleteMin : Option[Int] = getEpoch match{
    case Some(startEpoch) =>
      backoff.reset
      if(numMinsDelMin > 1) multiMinDeleteMin(startEpoch)
      else deleteMin1(startEpoch)
    case None => None
  }

  /** Delete the minimum node in the heap.
    * This version records a single minimal element while traversing. */
  private def deleteMin1(startEpoch: Long) : Option[Int] = {
    /* var retries = 0; */ var result : Option[Int] = null
    while(result == null){
      //retries += 1;
      // myAssert(retries<2*MaxRetries, "deleteMin retries exceeded: "+retries+"\n")
      val (pred, delNode) = findMinNode(startEpoch)
      if(delNode == head){ assert(pred == null); result = None }
      else if(tryDelete(pred, delNode, startEpoch)){
        maybeTidy(); result = Some(delNode.key)
      }
      else if(allowUnion && epoch.get != startEpoch) result = None
      // else retry
    }
    result
  }

  /** Find the node with the minimum key, and its predecessor.
    * @param startEpoch the epoch during which this call to deleteMin started. 
    * @return a triple (predMin, minNode) such that minNode held
    * the minimum key at some point during the search, and predMin was its
    * predecessor at some point.  Returns (null, head) if the heap is empty.
    */
  private def findMinNode(startEpoch: Long) : (Node, Node) = {
    var curr = head; var currSeq = curr.getSeq // current node and seq no
    var minNode = head; var predMin : Node = null // min node so far, and pred
    var minKey = Int.MaxValue
    while(curr != null) advance(curr, currSeq) match{
      case null =>  // restart
        backoff(); minNode = head; predMin = null
        if(allowUnion && epoch.get != startEpoch){
          // log("findMinNode epoch changed")
          curr = null // return (null, head)
        }
        else{ curr = head; currSeq = curr.getSeq; minKey = Int.MaxValue }
      case (next, nextSeq, skipNodes) =>
        val nextState = if(next == null) null else next.getState
        myAssert(next == null || nextState.seq != nextSeq ||
                   !nextState.label.isInstanceOf[Delete],
                 nextState.toString(2,1,next))
        val pred = curr; curr = next; currSeq = nextSeq
        if(curr != null && curr.key <= minKey){
          minNode = curr; minKey = curr.key
          if(skipNodes.isEmpty) predMin = pred else predMin = skipNodes.last
        }
        for(skipNode <- skipNodes; if skipNode.key <= minKey){
          val skipState = skipNode.getState
          skipState.label match{
            case Delete(prev) => // see if any child is new min
              var p = prev; var cs = skipState.children
              while(cs.nonEmpty){
                val c = cs.head; cs = cs.tail
                if(c.key <= minKey){ minNode = c; predMin = p; minKey = c.key }
                p = c
              }
            case _ => predMin = pred; minNode = skipNode; minKey = skipNode.key
          } // end of skipState.label match
        }
    }
    (predMin, minNode)
  }

  /** Version of deleteMin that stores up to numMins minimum nodes. */
  private def multiMinDeleteMin(startEpoch: Long): Option[Int] = /* Profiler.time("delMin") */ {
    // log("multiMinDeleteMin")
    val minList = new MinList[(Node,Node)](
      numMinsDelMin, _._2.key, (p1,p2) => p1._2 == p2._2)
    // Insert the children of deleted node and their predecessors, into
    // minList.  pred is predecessor of deleted node.
    def insertChildren(pred: Node, children: List[Node]) = {
      var cs = children; var p = pred
      while(cs.nonEmpty){
        val c = cs.head; cs = cs.tail; minList.insert(p, c); p = c
      }
    }
    // var iters = 0; val Limit = 100000
    var curr = head; var currSeq = curr.getSeq // current node and seq. no.
    var result : Option[Int] = null // the result to return, if non-null
    var deleteAttempts = 0

    // Restart a traversal
    def restart() = {
      if(epoch.get != startEpoch) result = None
      else{
        backoff(); curr = head; currSeq = curr.getSeq; minList.clear
      }
    }

    while(result == null){
      // Profiler.count("deleteMin iter")
      // iters += 1
      // if(iters > Limit-20) theLog.println(ThreadID.get+": "+curr.toString(2))
      // myAssert(iters < Limit, curr.toString(5))
      advance(curr, currSeq) match{
        case null => restart()
        case (next, nextSeq, skipNodes) =>
          // if(next != null){
          //   val nextState = next.getState
          //   if(nextState.seq == nextSeq)
          //     assert(nextState.parent == null ||
          //              nextState.parent.getLabel.isInstanceOf[Delete])
          // }
          var pred = curr // node before current one being considered
          for(skipNode <- skipNodes){
            val skipState = skipNode.getState
            if(skipState.label.isInstanceOf[Delete]){
              val children = skipState.children
              insertChildren(pred, children)
              if(children.nonEmpty) pred = children.last
            }
            else{ minList.insert(pred, skipNode); pred = skipNode }
          } // end of for loop
          if(next != null){
            minList.insert(pred, next)
            curr = next; currSeq = nextSeq // and continue
          }
          // else at end of list
          else if(minList.isEmpty){ assert(curr == head); result = None }
          else{
            val minPair = minList.get
            if(minPair == null) restart 
            else{
              deleteAttempts += 1
              val (predMin, delNode) = minPair
              // if(deleteAttempts > 10)
              //   log("Attempting to delete "+delNode.toString(2)+
              //         "; "+this.toString(2))
              if(tryDelete(predMin, delNode, startEpoch)){
                maybeTidy(); result = Some(delNode.key)
              }
              else{
                val delState = delNode.getState
                delState.label match{
                  case Delete(prev) =>
                    minList.removeFirst
                    insertChildren(prev, delState.children)
                  case _ => restart() // I think backtracking is not lock-free
                }
              } // end of inner else; then go back round the loop
            }
          } // end of outer else
      } // end of advance(...) match
    }
    // Profiler.count("delMin restart", restarts)
    // Profiler.count("tryDelete", deleteAttempts)
    // Profiler.count("delMin iters", iters)
    result
  }

  // ====================== Deletion

  /** Try to delete delNode with expected predecessor pred.  Test whether
    * relevant conditions hold.
    * @param startEpoch the epoch in which the deleteMin started. */
  private def tryDelete(pred: Node, delNode: Node, startEpoch: Long)
      : Boolean = {
    // Profiler.count("tryDelete")
    val delState = delNode.maybeClearParent(); val label = delState.label
    if(allowUnion && epoch.get != startEpoch){
      // log("tryDelete; epoch changed")
      false
    }
    else if(label != null){
      if(label.isInstanceOf[Delete])
        false // no point helping; fast return in this case
      else{ help(delNode, delState); tryDelete(pred, delNode, startEpoch) }
    }
    else{
      // Most calls that get here lead to a call to newDelete
      val parent = delState.parent
      if(parent != null){
        val pState = parent.getState
        pState.label match{
          case Delete(pPred) =>
            deleteWithParent(pred, delNode, delState, pPred,
                             parent, pState, startEpoch) ||
              tryDelete(pred, delNode, startEpoch)
          case _ => false  // minNode not a root, so retry
        }
      }
      else if(delNode.getState == delState){    // Try to perform labelling
        delete(pred, delNode, delState, startEpoch) ||
          tryDelete(pred, delNode, startEpoch)
      }
      else tryDelete(pred, delNode, startEpoch)
    }
  }


}
