package binomialHeap.concurrent

import java.util.concurrent.atomic.AtomicReference

/** A label encapsulating a list of HeapLabels and at most one UnionReceiver
  * label. */
case class HeapLabelList private (heap: ConcHeapBase) extends Label{

  /** The current UnionReceiver, possibly null; and the dummy header of the list
    * of HeapLabels. */
  private val state = new AtomicReference[(UnionReceiver, HeapLabel)](null)

  /** The initial dummy header. */
  private val head = new HeapLabel

  /** A reference to the last node in the list of HeapLabels.  */
  private val tail = new AtomicReference(head)

  /* The HeapLabels are formed into a linked list, with dummy header state._2.
   * If the last node points back to itself, then no new labels may be
   * added. */

  /** Auxilliary constructor, taking a UnionReceiver label. */
  def this(heap: ConcHeapBase, ur: UnionReceiver) = {
    this(heap); state.set(ur, head)
  }

  /** Auxilliary constructor, taking a HeapLabel. */
  def this(heap: ConcHeapBase, hl: HeapLabel) = {
    this(heap); head.next.set(hl); tail.set(hl); state.set(null, head)
  }

  /** This variable is set to true to indicate that a thread is trying to close
    * this HeapLabelList, to replace it with another.  Threads that see this
    * should not add more labels to this, but instead should help to close
    * this. */
  @volatile private var closing = false

  override def toString = {
    var st = "HeapLabelList("+heap.id+"): "
    val pair = state.get
    if(pair != null){
      // Traverse list using n, adding information about each node onto sts
      val (ul, h) = pair;
      st += (if(ul != null) ul.toString else "null")+"; "
      var n = h.next.get; var sts = List[String]()
      var done = (n eq h) || n == null
      while(!done){
        sts = n.toString :: sts
        val next = n.next.get
        if(next == null || (next eq n)) done = true
        else n = next
      }
      st+sts.reverse.mkString("<", ", ", ">")
    }
    else st+"<>"
  }

  /** Add hl to the list. 
    * @return false if this has been closed for new additions. */
  def add(hl: HeapLabel): Boolean = {
    var maxDelay = 3200
    if(closing){// this is being closed; help with that
      // heap.log("addHL helping to close")
      close1; false
    }
    else{
      val last = tail.get; val next = last.next.get
      if(next == null){ // try to append hl after last
        if(last.next.get == next && last.next.compareAndSet(null, hl)){
          tail.compareAndSet(last, hl); true
        }
        else{
          ox.cads.util.NanoSpin(heap.random.nextInt(maxDelay)) // Backoff
          maxDelay = (2*maxDelay) min 100000
          add(hl) // retry
        }
      }
      else if(next eq last){ // this is being closed; help with that
        close2; false
      }
      else{  // help with previous enqueue then re-try
        ox.cads.util.NanoSpin(heap.random.nextInt(maxDelay)) // Backoff
        maxDelay = (2*maxDelay) min 100000
        if(tail.get == last) tail.compareAndSet(last, next)
        add(hl)
      }
    }
  }

  /** Remove done HeapLabels from the front of the list. */
  def tidy = if(!closing){
    val myState = state.get
    if(myState != null){
      val (ur,head) = myState
      if(false){ // This version removes just from the front of the list
        var curr = head; var next = curr.next.get 
        // Inv: all nodes up to curr are done; next = curr.next.get
        while(next != null && next != curr && next.done){
          curr = next; next = curr.next.get
        }
        if(curr != head && state.get == myState)
          state.compareAndSet(myState, (ur, curr))
      }
      else{ // Try to remove all done nodes, excluding the tail
        var prev = head; var done = false
        // Inv: we've traversed up to prev (inclusive); prev != null
        do{
          val first = prev.next.get // store for CAS below
          var curr = prev; var next = first
          // Inv: all nodes from prev.next = first (inclusive) up to next
          // (exclusive) are done; next = curr.next.get
          while(next != null && next != curr && next.done){
            curr = next; next = curr.next.get
          }
          if(curr != prev && prev.next.get == first)
            prev.next.compareAndSet(first, if(next == null) curr else next)
          if(next == null || next == curr) done = true
          else prev = next
        } while(!done)
      }
    }
  }
  // Note: there seems little difference in performance between the two
  // versions, above.

  /** Add ur to this.  
    * @return true if successful or if another thread added ur (including if
    * it has subsequently been removed). */
  def add(ur: UnionReceiver): Boolean = {
    if(closing){
      // heap.log("addUR helping to close(1)")
      close1; false
    }
    else{
      val myState = state.get
      if(myState == null){
        // heap.log("addUR helping to close(3)")
        close3; false
      }
      else{
        val (ur1, head) = myState
        if((ur1 eq ur) || ur.done){
          // heap.log("Another thread added "+ur+" to HeapLabelList "+ur.done)
          true
        }
        else{
          if(ur1 != null && !ur1.done){ // need to help here
            // heap.log("Helping with "+ur1)
            heap.helpUnionReceiver(ur1); assert(ur1.done) 
          }
          state.get == myState && state.compareAndSet(myState, (ur, head)) ||
            add(ur) // retry if CAS unsuccessful
        }
      }
    }
  }

  /** Get the current UnionReceiver, possibly null. */
  def getUR: UnionReceiver = state.get match{
    case null => null
    case (ur, _) => ur
  }

  /** Remove the current UnionReceiver if it is ur. */
  def remove(ur: UnionReceiver): Unit = {
    assert(ur.done)
    val myState = state.get
    if(myState != null){
      val (ur1, head) = myState
      if((ur1 eq ur) && !state.compareAndSet(myState, (null, head)))
        remove(ur)
    }
  }

  /** Prevent more labels from being added to this.  Help to complete all the
    * existing labels.  Then remove this from head.  */
  def close = { closing = true; close1 }

  /** Help to complete all the labels in the queue, and set head.next = head to
    * prevent more from being added. */
  private def close1: Unit = {
    assert(closing)
    val theState = state.get
    if(theState == null) close3
    else{
      val (ur, head) = theState
      val first = head.next.get
      if(first == null){
        // all heap labels completed; close the loop on head to prevent more
        // from being added.
        if(head.next.compareAndSet(null, head)) close2 else close1
      }
      else if(first eq head) close2  // loop already closed
      else{
        if(!first.done){ heap.helpHL(first); assert(first.done) }
        // try to remove first from the list
        state.compareAndSet(theState, (ur, first)); close1
      }
    }
  }

  /** Help to complete ur, then remove this label.  
    * Pre: the linked list is empty, and header.next eq header. */
  private def close2 = {
    var done = false
    while(!done){
      val myState = state.get
      if(myState == null) done = true //  another thread has cleared the state
      else{
        val (ur, header) = myState
        assert(header.next.get eq header)
        if(ur != null && !ur.done){ // help to complete ur
          heap.helpUnionReceiver(ur); assert(ur.done)
        } 
        done = state.get == myState && state.compareAndSet(myState, null)
      }
    } // end of while(!done)
    close3
  }

  /** Complete closing of this, by removing this from heap head.
    * Pre: the state has been set to null.  */
  private def close3 = {
    var done = false
    while(!done){
      val heapHead = heap.head; val headState = heapHead.getState
      done = (headState.label ne this) ||
        heapHead.compareAndSet(headState, headState.addLabel(null))
    }
    // heap.log("Removed HeapLabelList: "+this)
  }
}

