package binomialHeap.tests

/** A linearizability tester for binomial heaps. */

import ox.cads.testing._
import binomialHeap.concurrent._ 
import binomialHeap.PriorityQueue
import ox.cads.util.{Profiler,ThreadID}

object HeapTest{
  var iters = 100  // # iterations in each run
  var reps = 1000  // # repetitions
  var p = 4
  val MaxKey = 5000
  var useLog = false // ; var logRoots = 0
  var useSkiplist = false; var useLJSkiplist = false
  var numMinsMin = 1; var numMinsDelMin = 1;
  var tidyRatio = 8
  var backtracking = true; var retrieving = false

  // val insertProb = 0.60; val minProb = 0.00; val delMinProb = 0.40
  val insertProb = 0.40; val minProb = 0.30; val delMinProb = 0.30
  assert(insertProb+minProb+delMinProb == 1.0)  

  /** A worker for the linearizability tester. */
  def worker(me: Int, log: GenericThreadLog[SeqMultiSet, PriorityQueue]) = {
    val random = new scala.util.Random(me+scala.util.Random.nextInt)
    for(_ <- 0 until iters){
      val rand = random.nextFloat
      if(rand <= insertProb){ 
        val x = random.nextInt(MaxKey)
        log(_.insert(x), "insert "+x, _.insert(x))
      }
      else if (rand <= minProb+insertProb)
        log(_.minimum, "minimum", _.minimum)
      else log(_.deleteMin, "delete minimum", _.deleteMin) 
    }
  }

  /** Run linearizability tests on a single binomial heap. */
  def doTests = {
    val dotStep = if(useLog) 1 max (4000/p/iters)  else 1 max (40000/p/iters)
    var preJIT = true

    // Now run the tests
    for(i <- 0 until reps){
      val logName = if(useLog) "/tmp/log" else ""
      val theLog =
        if(useLog) new ox.cads.util.TSLog(logName, p)
        else ox.cads.util.NullLog
      val concHeap : binomialHeap.PriorityQueue =
        if(useSkiplist) new binomialHeap.skiplist.SLHeap(logName)
        else if(useLJSkiplist) new binomialHeap.skiplist.LJPQueue()
        else new ConcHeap(
          theLog = theLog, MaxRetries = 200 max (if(preJIT) p*45 else p*15),
          numMinsMin = numMinsMin, numMinsDelMin = numMinsDelMin,
          tidyRatio = tidyRatio, backtracking = backtracking,
          retrieving = retrieving)
      val seqHeap = new SeqMultiSet
      val tester =
        LinearizabilityTester.JITGraph[SeqMultiSet, binomialHeap.PriorityQueue](
          seqHeap, concHeap, p, worker, iters, tsLog = true)
      var result = -1
      //result = tester()
      //if(result <= 0){ concHeap.theLog.shutdown ; sys.exit }
      try{ result = tester() } finally{ concHeap.shutdown }
      assert(result > 0) 
      if(i%dotStep == 0) print(".")
      if(i == 4*dotStep){
        Profiler.clear; println("Cleared Profiler"); preJIT = false
      }
      if(i%(1000*dotStep) == 0) println("\n"+i)
    }
    println
    // if(logRoots == 1) produceRootsStats() else  Profiler.report
  }

  // --------- Multi-heap linearizability testing
  val N = 3 // Number of heaps

  /** A worker for the multi-heap linearizability tester. */
  def multiHeapWorker(
      me: Int, log: GenericThreadLog[SeqMultiHeap, MultiHeap]) = {
    val random = new scala.util.Random(me+scala.util.Random.nextInt)
    for(_ <- 0 until iters){
      val i = random.nextInt(N) // heap to use
      val rand = random.nextFloat
      if(rand <= 0.2){ 
        val j = (i+1+random.nextInt(N-1))%N; assert(i != j)
        // println(ThreadID.get+": union"+(i,j))
        log(_.union(i, j), "union("+i+", "+j+")", _.union(i, j))
      }
      else if(rand <= 0.5){
        //println("insert")
        val x = random.nextInt(MaxKey)
        log(_.insert(i, x), "insert("+i+", "+x+")", _.insert(i,x))
      }
      else if (rand <= 0.8)
        log(_.minimum(i), "minimum("+i+")", _.minimum(i))
      else log(_.deleteMin(i), "deleteMin("+i+")", _.deleteMin(i))
    }
  }

  /** Run linearizability tests on a multi-heap.
    * iters needs to be quite small, say 20, for the linearization testing
    * to work.
    * @param concUnion use the concurrent union algorithm. */
  def doMultiHeapTests(concUnion: Boolean) = {
    println("doMulitiHeapTests: "+N)
    val dotStep = if(useLog) 1 max (2000/p/iters)  else 1 max (20000/p/iters)
    var preJIT = true

    // Now run the tests
    for(i <- 0 until reps){
      val logName = if(useLog) "/tmp/log" else ""
      val theLog =
        // if(useLog) ox.cads.util.ScreenLog 
        if(useLog) new ox.cads.util.TSLog(logName, p)
        else ox.cads.util.NullLog
      val concHeaps =
        if(concUnion)
          new ConcurrentMultiHeap(
            theLog, N, numMinsMin = numMinsMin, numMinsDelMin = numMinsDelMin)
        else new LockedMultiHeap(theLog, N)
      val seqHeaps = SeqMultiHeap(N)
      val tester =
        LinearizabilityTester.JITGraph[SeqMultiHeap, MultiHeap](
          seqHeaps, concHeaps, p, multiHeapWorker, iters, tsLog = true)
      try{
        val result = tester(); theLog.assert(result > 0)
      } finally { theLog.shutdown }
      if(i%dotStep == 0) print(".")
      if(i == 4*dotStep){
        Profiler.clear; println("Cleared Profiler"); preJIT = false
      }
      if(i%(1000*dotStep) == 0) println("\n"+i)
    }
    println
    // if(logRoots == 1) produceRootsStats() else  Profiler.report
  }

  // =======================================================

  /** Produce statistics on the number of roots. */
  def produceRootsStats() = {
    val stats: List[(String,Long)] = Profiler.getReport._2
    // Each entry above is of the form, e.g. (" 14,  3", 5); get the
    // corresponding triple (14, 3, 5), representing (#nodes, #roots, count).
    def getPair(pair: (String, Long)) : (Int, Int, Long) = {
      val (st, n) = pair; val a = st.split(","); assert(a.length == 2)
      (a(0).dropWhile(_ == ' ').toInt, a(1).dropWhile(_ == ' ').toInt, n)
    }
    val stats1 = stats.map(getPair)
    val maxNodes = stats1.map(_._1).max
    // Average #roots for a particular group
    def average(group: List[(Int, Int, Long)]) : Float =
      if(group.nonEmpty)
        group.map{case (n,r,c) => r*c}.sum.toFloat / group.map(_._3).sum
      else 0.0F
    val averages =
      Array.tabulate(maxNodes+1)(n => (n, average(stats1.filter(_._1 == n))))
    println(stats.map(getPair).mkString("\n"))
    println(averages.mkString("\n"))
    println("Maximum roots = "+stats1.map(_._2).max)
    println("Minimum roots = "+stats1.map(_._2).min)
  }

  // ---------

  def main(args: Array[String]) = {
    /** Parse arguments. */
    var i = 0; var useMultiHeaps = false; var concUnion = false
    while(i < args.length) args(i) match{
      case "-p" => p = args(i+1).toInt; i += 2 
      case "--iters" => iters = args(i+1).toInt; i += 2 
      case "--reps" => reps = args(i+1).toInt; i += 2 
      case "--useLog" => useLog = true; i += 1 
      // case "--logRoots" => logRoots = 1; i += 1 
      case "--skiplist" => useSkiplist = true; i += 1
      case "--LJskiplist" => useLJSkiplist = true; i += 1
      case "--numMinsMin" => numMinsMin = args(i+1).toInt; i += 2 
      case "--numMinsDelMin" => numMinsDelMin = args(i+1).toInt; i += 2 
      case "--tidyRatio" => tidyRatio = args(i+1).toInt; i += 2
      case "--multiHeaps" => useMultiHeaps = true; i += 1
      case "--concUnion" => concUnion = true; i += 1
      case _ => println("Unrecognised argument: "+args(i)); sys.exit
    }
    Profiler.setWorkers(p)
    if(useMultiHeaps) doMultiHeapTests(concUnion) else doTests
  } 


}
