package binomialHeap.skiplist

import java.util.concurrent.atomic._
import ox.cads.atomic.AtomicPair
import ox.cads.util.ThreadID

/** A concurrent priority queue implemented using a skiplist.  Based on "A
  * Skiplist-Based Concurrent Priority Queue with Minimal Memory Contention",
  * by Linden and Jonnson. */

/** A node in the LJ skiplist. */
class LJNode(val key: Int, val level: Int){
  /** Is this node being inserted? */
  var inserting = true

  /** The next element of the skiplist, for levels >0. */
  val next = new AtomicReferenceArray[LJNode](level)

  /** Atomic pair giving a pointer to the next node at level 0, and flag
    * indicating whether that node is deleted. */
  val pair = new AtomicPair[LJNode, Boolean](null, false)

  def next0 = pair.getFirst // next at level 0
  def d = pair.getSecond // is the next node deleted?

  /** Get next node at level l. */
  def getNext(l: Int) = if(l == 0) next0 else next.get(l)

  /** The pair (next(l), d), but from a consistent state. */ 
  def getPair(l: Int) : (LJNode, Boolean) = {
    // The following uses the fact that d never changes from true to false.
    val d1 = d; val next = getNext(l); 
    if(d == d1) (next, d1)
    else{ assert(d); (getNext(l), true) } //  getPair(l)
  }
}

// ==================================================================

class LJPQueue(boundOffset: Int = 20) extends binomialHeap.PriorityQueue{
  /** Maximum number of levels in any node. */
  private val MaxLevel = 32

  private val TopLevel = MaxLevel-1 // top valid level

  // private val BoundOffset = 20

  /** Sentinel head node. */
  private val head = new LJNode(Int.MinValue, MaxLevel)
  /** Sentinel tail node. */
  private val tail = new LJNode(Int.MaxValue, MaxLevel)

  // initialise head's pointers to point to tail
  for(i <- 1 until MaxLevel) head.next.set(i, tail)
  head.pair.set((tail, false))
  head.inserting = false; tail.inserting = false

  def deleteMin() : Option[Int] = {
    var x = head; var offset = 0; var newHead: LJNode = null;
    var oldHead = x.next0; var done = false

    do{
      val (nxt, d) = x.pair.get
      if(nxt == tail) return None // queue empty
      if(x.inserting && newHead == null) newHead = x
      if(!d){
        if(x.pair.compareAndSet((nxt, false), (nxt, true))){
          offset += 1; x = x.next0; done = true
        }
        // else continue loop with same x
      }
      else{ offset += 1; x = x.next0 }
    } while(!done)

    if(offset >= boundOffset){
      if(newHead == null) newHead = x
      if(head.pair.compareAndSet((oldHead, true), (newHead, true)))
        restructure()
    }
    Some(x.key)  
  }

  /** Rebuild the heap structure after deletion. */
  private def restructure() = {
    var i = TopLevel; var h = head.getNext(i); var curr = h
    while(i > 0){
      if(curr == tail) i -= 1
      else{
        while(curr.d) curr = curr.getNext(i)
        val success = head.next.compareAndSet(i, h, curr)
        if(success) i -= 1
        h = head.getNext(i)
        if(!success) curr = h
      }
    }

  }

  /** Get the minimum key. */
  def minimum: Option[Int] = {
    var x = head; var done = false; var next : LJNode = null
    do{
      val (nxt, d) = x.pair.get; next = nxt
      if(nxt == tail || !d) done = true
      else x = x.next0
    }while(!done);
    if(next == tail) None else Some(next.key)
  }

  /** Insert k. */
  def insert(k: Int) = {
    val height = randomLevel; val newNode = new LJNode(k, height)
    var done = false; var skew = false
    var preds: Array[LJNode] = null; var succs: Array[LJNode] = null
    do{
      val (preds1, succs1, skew1) = locatePreds(k)
      preds = preds1; succs = succs1; skew = skew1
      newNode.pair.set((succs(0), false))
      done = preds(0).pair.compareAndSet((succs(0), false), (newNode, false))
      // if(!done) println(ThreadID.get+" "+k+" "+preds(0).key+" "+preds(0).d+" "+succs(0).key)
    }while(!done);
    // insert at higher levels
    var i = 1; done = false
    while(i < height && !skew && !done){
      newNode.next.set(i, succs(i))
      if(newNode.d) done = true
      else if(preds(i).next.compareAndSet(i, succs(i), newNode))
        i += 1 // inserted at this level
      else{
        val (preds1, succs1, skew1) = locatePreds(k)
        preds = preds1; succs = succs1; skew = skew1
        if(succs(0) != newNode) done = true
      }
    }
    newNode.inserting = false
  }

  /** Pick a random level at which to insert a new node. */
  private def randomLevel: Int = {
    var l = 1
    while(scala.util.Random.nextBoolean && l < MaxLevel) l += 1
    l
  }

  /** Locate the predecessors of a node for k.
    * @return a triples (preds, succs, skew), where k should be inserted
    * between preds(i) and succs(i) at each level i, and ... */
  private def locatePreds(k: Int)
      : (Array[LJNode], Array[LJNode], Boolean) = {
    val preds, succs = new Array[LJNode](MaxLevel)
    var i = TopLevel; var skew = false; var pred = head
    while(i >= 0){
      val (curr1, d1) = pred.getPair(i); var curr = curr1; var d = d1
      // var curr = pred.getNext(i); var d = pred.d // atomic??
      while(curr.key < k || curr.d || (d && i == 0)){
        if(i < TopLevel && curr == succs(i+1)) skew = true
        pred = curr // ; curr = pred.getNext(i); d = pred.d
        val (curr1, d1) = pred.getPair(i); curr = curr1; d = d1
      }
      preds(i) = pred; succs(i) = curr; i -= 1
    }
    (preds, succs, skew)
  }

  def shutdown: Unit = { }

  def union(giver: binomialHeap.concurrent.Union) = ???

}
