package binomialHeap.concurrent

import java.util.concurrent.atomic.{AtomicReference,AtomicInteger}

/** Labels that can be added to Nodes. */
abstract class Label{
  /** Equality test, overridden as object equality. */
  override def equals(that: Any) = that match{
    case label: Label =>  this eq label
    case null => false
  }
  // However, I'll try to use "eq" or "ne" in code.  And the above is probably
  // overwritten by case classes.
}

/** Delete the marked node, promoting its children to root nodes. 
  * @param prev the previous node
  * @param children the children of the marked node, in order. 
  * if the marked node is distinct from advance.next and advance.parent.
  */
case class Delete(prev: Node) extends Label{
  val ts = java.lang.System.nanoTime

  override def toString = "Delete{prev = "+prev.toString(0,0,prev)+"}"

  @volatile var heap: Deletion = null

  /** Equality test, overridden as object equality. */
  override def equals(that: Any) = that match{
    case label: Label =>  this eq label
    case null => false
  }
}

/** Merge this node (t1) and t2, with expected state state2, making t2 a child
  * of this.  pred is the expected predecessor of t2, with expected state
  * pState, and pLabel is an label to be added to it.  */
case class MergeParent(
  pred: Node, predState: NodeState, pLabel: MergeNext,
  t2: Node, state2: NodeState)
    extends Label{
  override def toString = "MergeParent"+(pred.key, t2.key)
}


/** Merge t1 and t2, with expected state state2, making t2 a child of t1.
  * This labels the expected predecessor of t2. */
case class MergeNext(t1: Node, t2: Node, state2: NodeState) extends Label{
  override def toString = "MergeNext"+(t1.key, t2.key)

  /** Equality test, overridden as object equality. */
  override def equals(that: Any) = that match{
    case label: Label => this eq label
    case null => false
  }
}


/** This node has just been used in a union, as the last node of the receiving
  * heap. */
case class UnionLast(gStatus: UnionGiver) extends Label{
  override def toString = "UnionLast("+gStatus+")"
}

// ==================================================================

/** HeapStatus indicating that a thread is trying to perform a union
  * operation, with this as the giving heap, and receiver as the receiving
  * heap.  
  * @param ur is the UnionReceiver label for the receiver; inv: if this 
  * status is on heap g, then ur == UnionReceiver(g), and ur will be
  * put on receiver. */
case class UnionGiver(receiver: Union, ur: UnionReceiver) extends Label{
  /** A triple of the first node, the heap indirection and the epoch of the
    * giving heap.  Initialised by init. */
  private val state = new AtomicReference[(Node, HeapIndirection, Long)](null)

  /** Initialise with the first node, the heap indirection and the epoch of the
    * giving heap. */
  def init(first: Node, hi: HeapIndirection, epoch: Long) = {
    if(state.get != null || ! state.compareAndSet(null, (first, hi, epoch)))
      assert(state.get == (first, hi, epoch) || joined || aborted)
  }

  def isInit = state.get != null 

  // The first node of the giver heap. 
  def first = state.get._1

  /** The HeapIndirection of the giving heap. */
  def heapIndirection = state.get._2

  /** The epoch of the giving heap. */
  def epoch = state.get._3

  /** If the labelling using this is aborted, the UnionReceiver label that
    * replaces it. */
  @volatile var replacedBy: UnionReceiver = null

  /** Abort this labelling, setting ur to be the UnionReceiver label that
    * replaces it. */
  def abort(ur: UnionReceiver) = {
    assert(replacedBy == null || (replacedBy eq ur))
    replacedBy = ur
  }

  /** Has the labelling with this been aborted? */
  def aborted = replacedBy != null

  /** A sequence counter; unique to this heap (ignoring overflow). */
  val seq = UnionGiver.seq.getAndIncrement

  override def toString = "UnionGiver("+receiver.id+", "+seq+", "+ur+")"

  /** A reference to a triple maybe giving information about the last node in
    * the receiving heap: that node, its state, and a boolean.  The former two
    * may be null, if unknown.  The boolean indicates whether the linearization
    * point has been reached, i.e. the last node has been updated to point to
    * the giving heap and a UnionLast label added.  If the boolean is false,
    * the former two may be updated. */
  private val lastInfo =
    new AtomicReference[(Node, NodeState, Boolean)](null, null, false)

  /** Get information about the last node in the receiving heap: the last node
    * (or null, if unknown), its state, and whether the linearization point of
    * the union has been reached.  */
  def getLastInfo: (Node, NodeState, Boolean) = lastInfo.get
 
  /** Try to set information about last node of the receiving heap, if not
    * already set.
    * @param last a recent last node
    * @param lastState the last node's state
    * @return true if successful or another thread has made the change. */
  def setLastInfo(last: Node, lastState: NodeState): Boolean = {
    val oldLastInfo @ (oldLast, oldLastState, done) = lastInfo.get
    // val (oldLast, oldLastState, done) = oldLastInfo
    oldLast == null &&
      lastInfo.compareAndSet(oldLastInfo, (last, lastState, false)) ||
      // test if another thread made change, but not yet done
      oldLast == last && oldLastState == lastState && !done 
  } 

  /** Clear the information stored about the last node in the receiving heap, if
    * it corresponds to last and lastState.  This should be called only if the
    * state of last has changed from lastState other than to join the
    * heaps. */
  def clearLastInfo(last: Node, lastState: NodeState) = {
    val oldLastInfo @ (oldLast, oldLastState, done) = lastInfo.get
    // val (oldLast, oldLastState, done) = oldLastInfo
    if(last == oldLast && lastState == oldLastState){
      assert(!done); lastInfo.compareAndSet(oldLastInfo, (null, null, false))
    }
  }

  /** Indicate that the joining CAS has been performed from node last, giving
    * state newLastState for last. */
  def setDone(last: Node, newLastState: NodeState) = {
    val oldLastInfo @ (oldLast, oldLastState, done) = lastInfo.get
    // The first disjunct below is necessary: joined could call setDone with
    // an old value for last that happens now to point to first (after the
    // joining CAS).
    assert(done || last == oldLast)
    if(!done) lastInfo.compareAndSet(oldLastInfo, (last, newLastState, true))
  }

  /** Has the joining CAS been performed (so that the union is linearized)? */
  def joined = {
    val (last, lastState, d) = lastInfo.get
    if(d) true
    else if(last != null){
      val newLastState = last.getState
      if(newLastState.next == first){ setDone(last, newLastState); true }
      // The following is necessary in case the joining CAS happened before
      // the call to done, but the label had been cleared before the call to
      // last.getState (in which case setDone would have been called between
      // these two events).
      else lastInfo.get._3
    }
    else false
  }
}

/** Companion object for UnionGiver. */
object UnionGiver{
  // A sequence counter. 
  val seq = new AtomicInteger(0)
}

// ==================================================================

/** A label indicating that node has just been inserted after this node, with
  * head labelled by ia. */
case class Inserted(node: Node, ia: InsertAtEnd) extends Label{
  override def toString = "Inserted"+(node.key, ia)
} 

// ==================================================================

