package binomialHeap.concurrent

import ox.cads.util.Profiler

/** A list of the minimal nodes found in current traversals.  In minimum, the
  * type N is Node, and getKey returns the node's key.  In deleteMin, the type
  * T is (Node, Node), representing a pair of adjacent nodes, and getKey
  * returns the latter's key. */
trait MinListT[N]{
  /** Insert node into this. */
  def insert(node: N) : Unit

  /** Remove the first node, which has been deleted. */
  def removeFirst() : Unit

  /** Clear this. */
  def clear : Unit

  /** Get the minimum element, or null if this is empty. */
  def get : N

  /** Is the list empty. */
  def isEmpty : Boolean
}

// =======================================================

/** An implementation of MinListT to hold up to NumMins values, using an
  * array.  
  * 
  * This implementation assumes that the NumMins+1 smallest values are not all
  * equal to Int.MaxValue */
class MinList[N : scala.reflect.ClassTag](
  NumMins: Int, getKey: N => Int, matches: (N, N) => Boolean)
    extends MinListT[N]{
  // deleted: N => Boolean = (n: N) => false
  assert(NumMins >= 1)

  /** Array holding the minimal nodes inserted and not replaced so far, in
    * increasing order of key.  */
  private val minList = new Array[N](NumMins)

  /** Array holding the corresponding keys. */
  private val keys = new Array[Int](NumMins)

  /** Number of valid entries. */
  private var size = 0

  // Invariant: size <= NumMins && minList[0..size) are non-null and have
  // non-decreasing keys, and no repetitions && for all i in [0..size),
  // keys(i) = getKey(minList(i)).

  /** The minimal key for which there was no room in minList. */
  private var minDiscard = Int.MaxValue
  // Invariant minDiscard >= all keys in minList[0..size).
  // All other keys ever inserted are at least minDiscard. 

  // It is possible that there are nodes with keys greater than or equal to
  // minDiscard for which insert was called but for which are not in minList.
  // Thus subsequently nodes with keys greater than minDiscard should not be
  // stored.

  /** Insert node into this. */
  def insert(node: N) : Unit = {
    val key = getKey(node) 
    if(key <= minDiscard){
      if(size < NumMins || keys(size-1) > key){
        // Find where to insert node
        var i = 0
        while(i < size && keys(i) < key) i += 1
        // Test if node there already
        var j = i
        while(j < size && keys(j) == key && !matches(minList(j), node)) j += 1
        if(j < size && keys(j) == key){ // just overwrite
          // assert(matches(minList(j), node))
          minList(j) = node
        }
        // if(i < size && matches(minList(i), node)){ } // return; node already there
        else{
          // Need to store node into position store
          val store = i; i = size
          if(size == NumMins){ // will discard last element
            i = size-1; assert(keys(i) < Int.MaxValue); minDiscard = keys(i)
          }
          else size += 1
          // minList[0..i) ++ minList[i+1..size) holds original list;
          // minList[i+1..size).key > key
          while(i > store){
            minList(i) = minList(i-1); keys(i) = keys(i-1); i -= 1
          }
          minList(i) = node; keys(i) = key
        }
      } // end of if(size < NumMins || minList(size-1).key > key)
      else{ // discard node, but it's the smallest discarded.
        assert(key < Int.MaxValue); minDiscard = key
      }
    }
    // else key > minDiscard, so we shouldn't store node
  }

  /** Insert node into this, purging deleted nodes. */
  // This version is faulty: it is necessary to insert the children of any
  // inserted node.
  // def insertPurge(node: N) : Unit = {
  //   val key = getKey(node) 
  //   if(key <= minDiscard){
  //     // Find where to insert node
  //     var i = 0; var j = 0
  //     // The contents are in positions [0..i) ++ [j..size), with i <= j <=
  //     // size.  In the first loop, keys[0..i) < key.
  //     while(j < size && keys(j) < key){
  //       if(true || !deleted(minList(j))){
  //         if(i != j){ keys(i) = keys(j); minList(i) = minList(j) }
  //         i += 1
  //       }
  //       j += 1
  //     }
  //     // Insert node in position i
  //     if(i < j){
  //       if(minList(i) != node){ keys(i) = key; minList(i) = node; i += 1 }
  //       if(i < j){ // copy rest down, purging
  //         while(j < size){
  //           if(true || !deleted(minList(j)) && minList(j) != node){
  //             keys(i) = keys(j); minList(i) = minList(j); i += 1
  //           }
  //           j += 1
  //         }
  //         size = i
  //       } // end of inner if(i < j)
  //     } // end of outer if(i < j)
  //     else{
  //       assert(i == j && i <= size)
  //       if(i < size){   // Need to shuffle entries up to make room
  //         var copyN = minList(i); var copyK = keys(i)
  //         minList(i) = node; keys(i) = key; i += 1
  //         // The contents are in minList[0..i) ++ [copyN] ++ minList[i..size),
  //         // with matching keys, and with copyK = getKey(copyN).
  //         while(i < size /* && !deleted(copyN) */ && copyN != node){
  //           val tempN = minList(i); minList(i) = copyN; copyN = tempN
  //           val tempK = keys(i); keys(i) = copyK; copyK = tempK
  //           i += 1
  //         }
  //         if(!deleted(copyN) && copyN != node){
  //           assert(i == size)
  //           if(i < NumMins){ minList(i) = copyN; keys(i) = copyK; size += 1 }
  //           else{  // discard copyN, but it's the smallest discarded.
  //             assert(copyK <= minDiscard); minDiscard = copyK }
  //         }
  //         // else purge copyN
  //       } // end of if(i < size)
  //       else if(size < NumMins){
  //         minList(i) = node; keys(i) = key; size += 1
  //       }
  //       else{ assert(key <= minDiscard); minDiscard = key } // discard node
  //     }
  //   } // end of if(key <= minDiscard); else don't store node
  // } 

  /** Remove the first node, which has been deleted. */
  def removeFirst() = {
    // assert(minList(0).getLabel.isInstanceOf[Delete] && size > 0)
    assert(size > 0)
    var i = 1
    // Shuffle nodes down to overwrite deleted node
    while(i < size){ minList(i-1) = minList(i); keys(i-1) = keys(i); i += 1 }
    size -= 1
  }

  /** Clear this. */
  def clear = {
    // for(i <- 0 until NumMins) minList(i) = null.asInstanceOf[N]
    minDiscard = Int.MaxValue; size = 0
  }

  /** Get the minimum element. 
    * @return null if this is empty. */
  def get : N = 
    if(size == 0) null.asInstanceOf[N] else minList(0)
    //  assert(size > 0); minList(0) 

  /** Is the list empty. */
  def isEmpty = size == 0 && minDiscard == Int.MaxValue

  /** Clone this into that. */
  def cloneTo(that: MinList[N]) = {
    minList.copyToArray(that.minList, 0, size)
    keys.copyToArray(that.keys, 0, size)
    that.size = size; that.minDiscard = minDiscard
  }
}

// =======================================================

/** An implementation of MinListT to hold up to 2 values.
  * This is effectively an optimisation of MinList[N](2, getKey), to avoid 
  * using an array. */

class MinList2[N](getKey: N => Int)  extends MinListT[N]{
  /** The current entries (possibly invalid). */
  private var n0, n1 = null.asInstanceOf[N]

  /** The keys. */
  private var k0, k1 = -1
  // Invariant: if n0 != null, k0 = getKey(n0);
  // if n1 != null, k1 = getKey(n1) && k0 <= k1 && n0 != n1

  /** The number of valid entries. */
  private var size = 0
  // Inv: 0 <= size <= 2.  This represents the first size out of [n0, n1];
  // each of those is non-null.  Below, comments will sometimes pretend that
  // we are dealing with an array [n0, n1], and talk about positions in that
  // array.

  /** The minimal key for which there was no room. */
  private var minDiscard = Int.MaxValue
  // Invariant minDiscard >= k0, k1.

  def insert(node: N) : Unit = {
    val key = getKey(node) 
    if(key <= minDiscard){
      if(size == 2 && key >= k1)
        minDiscard = key // discard node, but it's the smallest discarded.
      else{
        // Store node in "position" store
        val store = if(size > 0 && key > k0) 1 else 0
        // Test if node already there
        if(store < size &&
             (store == 0 && n0 == node || store == 1 && n1 == node)){ } 
        else{
          if(size == 2) minDiscard = k1 // will discard last element
          else size += 1
          assert(store < size)
          if(store == 1){ n1 = node; k1 = key }
          else{ if(size == 2){ n1 = n0; k1 = k0 }; n0 = node; k0 = key }
        }
      }
    }
    // else key > minDiscard, so ignore
    if(size == 2) assert(k0 <= k1)
  }

  /** Remove the first node, which has been deleted. */
  def removeFirst() = {
    assert(size > 0)
    if(size == 2){ n0 = n1; k0 = k1; size = 1 }
    else{ assert(size == 1); size = 0 }
  }

  /** Clear this. */
  def clear = { minDiscard = Int.MaxValue; size = 0 }

  /** Get the minimum element.  */
  def get : N = { assert(size > 0); n0 }
  // if(size == 0) null.asInstanceOf[N] else

  /** Is the list empty. */
  def isEmpty = size == 0

  override def toString =
    if(size == 0) "<>" else if(size == 1) "<"+k0+">" else "<"+k0+","+k1+">"

}


// =======================================================

/** An unbounded implementation of MinListT, using a List.   */
class MinListUnbounded[N : scala.reflect.ClassTag](getKey: N => Int)
    extends MinListT[N]{

  /** List holding the nodes inserted so far, in increasing order of key, with
    * no repetitions. */
  private var minList = List[N]()

  /** Insert node into this. */
  def insert(node: N) : Unit = {
    val key = getKey(node)
    val (before, after)  = minList.span(getKey(_) < key)
    // Check if node already appears
    //if(before.filter(getKey(_) == key).forall(_ != node))
    minList = before ++ (node :: after)
  }

  /** Remove the first node, which has been deleted. */
  def removeFirst() = {
    assert(minList.nonEmpty); // minList = minList.tail
    val first = minList.head; val key = getKey(first)
    // remove duplicates of first
    val (sameKey, larger) = minList.tail.span(getKey(_) == key)
    minList = sameKey.filter(_ != first) ++ larger
  }

  /** Clear this. */
  def clear = minList = List()

  /** Get the minimum element. 
    * @return null if this is empty. */
  def get : N = if(minList.isEmpty) null.asInstanceOf[N] else minList.head

  /** Is the list empty. */
  def isEmpty = minList.isEmpty
}
