package binomialHeap.concurrent

import ox.cads.util.{Log,NullLog,ThreadID}

/** Adding the minimum operation to the binomial heap. */
abstract class Minimum(
  theLog: Log = NullLog, MaxRetries: Int = 2000,
  numMinsMin: Int = 1, numMinsDelMin: Int = 1,
  tidyRatio: Int = 8, RecTidy: Boolean = true,
  A: Int = 16, B: Int = 16,
  backtracking: Boolean = false, retrieving: Boolean = false,
  allowConcurrentUnion: Boolean = false)
    extends
    Traversing(theLog, MaxRetries, numMinsMin, numMinsDelMin, tidyRatio, 
                 RecTidy, A, B, backtracking, retrieving, allowConcurrentUnion){


  // ========= minimum

  /** Find the minimum key. */
  def minimum : Option[Int] = getEpoch match{
    case Some(startEpoch) =>
      if(numMinsMin > 1) multiMinimum(startEpoch) else minimum1(startEpoch)
    case None => None
  }

  /** Find the minimum key.
    * This version maintains a single minimum node while traversing. 
    * @return None if the heap is empty. */
  private def minimum1(startEpoch: Long) : Option[Int] = {
    //Profiler.count("minimum")

    var curr = head; var currSeq = curr.getSeq; // current node and seq. no
    var minNode = head; var minKey = Int.MaxValue // min node and key
    val startTime = java.lang.System.nanoTime
    // Update minNode and minKey, based on the nodes ns.
    def updateMin(ns: List[Node]) : Unit = for(n <- ns if n.key <= minKey){
      val nState = n.getState
      nState.label match{
        case delLabel: Delete
            if delLabel.ts - startTime <= 0 ||
                 allowConcurrentUnion && delLabel.heap != this =>
          updateMin(nState.children)
        case _ => minNode = n; minKey = n.key
      }
    }
    if(allowConcurrentUnion && epoch.get != startEpoch){
      // This is just an optimisation; the above test is cheaper than
      // epochValid
      // log("minimum1 restart; epoch of "+id+" changed")
      return None
    }
    while(curr != null){
      advance(curr, currSeq) match{
        case null => // restart
          curr = head; currSeq = curr.getSeq
          minNode = head; minKey = Int.MaxValue
          if(allowConcurrentUnion && epoch.get != startEpoch){
            // log("minimum1 restart; epoch of "+id+" changed")
            return None
          }
          // Profiler.count("minimum restart")
        case (next, nextSeq, skipNodes) =>
          val pred = curr; curr = next; currSeq = nextSeq 
          if(curr != null && curr.key <= minKey){
            minNode = curr; minKey = curr.key
          }
          if(true) updateMin(skipNodes)
          else for(skipNode <- skipNodes; if skipNode.key <= minKey){
            val skipState = skipNode.getState
            skipState.label match{
              case delLabel : Delete =>
                for(c <- skipState.children; if c.key <= minKey){
                  minNode = c; minKey = c.key
                }
              case _ => minNode = skipNode; minKey = skipNode.key
            }
          }
      }
    }
    if(minNode == head) None
    else if(minNode.deleted) minimum1(startEpoch)
    else if(allowConcurrentUnion && !epochValid(startEpoch)){
      // log("minimum1; epoch of "+id+" changed")
      None
    }
    else Some(minKey)
  }

  /** Alternative version of the minimum function, which keeps track of
    * numMinsMin minimal nodes while traversing. */
  private def multiMinimum(startEpoch: Long) : Option[Int] = {
    // log("multiMinimum")
    //Profiler.count("minimum")
    // var iters = 0; val Limit = 100000
    val minList : MinListT[Node] =
      new MinList[Node](numMinsMin, _.key, (n1,n2) => n1==n2)
    var curr = head; var currSeq = curr.getSeq;
    var result: Option[Int] =
      if(allowConcurrentUnion && epoch.get != startEpoch){
        // This is just an optimisation; the above test is cheaper than
        // epochValid
        // log("multiMinimum at start; epoch of "+id+" changed")
        None
      }
      else null
    val startTime = java.lang.System.nanoTime
    def restart() = {
      curr = head; currSeq = curr.getSeq; minList.clear
      if(allowConcurrentUnion && epoch.get != startEpoch){
        // This is just an optimisation; the above test is cheaper than
        // epochValid
        // log("multiMinimum restart; epoch of "+id+" changed")
        result = None
      }
      //Profiler.count("minimum restart")
    }
    // Insert the nodes ns into minList, except recursing if labelled for
    // deletion before startTime.  Return false if too many nodes were
    // encountered.
    def insertNodes(ns: List[Node]): Boolean = {
      var iters = 0; val Max = 200 // Number of iterations so far; max number
      // Note: Max needs to be greater than the number of threads,
      // for lock-freedom
      def insertNodes1(ns: List[Node]): Unit = {
        iters += 1
        if(iters > Max){ } // log("insertNodes giving up") 
        else{
          // assert(depth <= 18, ns.map(_.toString(2,2)).mkString("\n"))
          var ns1 = ns
          while(ns1.nonEmpty){
            val n = ns1.head; ns1 = ns1.tail
            val nState = n.getState
            nState.label match{
              case delLabel: Delete  =>
                if(delLabel.ts - startTime <= 0)
                  insertNodes1(nState.children)
                else if(allowConcurrentUnion){
                  assert(delLabel.heap != null)
                  if(delLabel.heap != this){
                    // log("insertNodes: node "+n.toString(1)+
                    //       " deleted from different heap "+delLabel.heap.id)
                    insertNodes1(nState.children)
                  }
                  else minList.insert(n)
                }
                else minList.insert(n)
              case _ => minList.insert(n)
            }
          } // end of while
        }
      }  // end of insertNodes1
      insertNodes1(ns); iters <= Max
    }
 
    while(result == null) advance(curr, currSeq) match{
      case null =>
        restart()
      case (next, nextSeq, skipNodes) => 
        // iters += 1
        // if(iters > Limit-20) theLog.println(ThreadID.get+": "+curr.toString(2))
        // myAssert(iters < Limit, curr.toString(5))
        val nextState = if(next == null) null else next.getState
        myAssert(next == null || nextState.seq != nextSeq ||
                   !nextState.label.isInstanceOf[Delete],
                 nextState.toString(2,1,next))
        if(!insertNodes(skipNodes)) restart()
        else if(next != null){
          minList.insert(next); curr = next; currSeq = nextSeq
        }
        else if(minList.isEmpty){ assert(curr == head); result = None }
        else{ // at end of list; find first non-deleted node in minList
          val minNode = minList.get
          if(minNode == null) restart()
          else if(minNode.deleted){ // minNode is fully deleted
            minList.removeFirst
            // insertNodes(minNode.getState.children)
            for(c <- minNode.getState.children) minList.insert(c)
            // no need for insertNodes here: minNode was not labelled for
            // deletion at the start of the call, so neither was any child.
          }
          else if(allowConcurrentUnion && !epochValid(startEpoch)){
            // log("multiMinimum; epoch of "+id+" changed")
            result = None
          }
          else result = Some(minNode.key)
        }
    } // end of while ... match
    result
  }

}
