package binomialHeap.tests

import binomialHeap.concurrent._
import ox.cads.util.{Log,NullLog}

/** A trait describing a collection of concurrent binomial heaps. */
trait MultiHeap{
  /** Insert key into heap i. */
  def insert(i: Int, key: Int) : Unit

  /** Minimum key in heap i. */
  def minimum(i: Int) : Option[Int]

  /** Delete minimum key in heap i. */
  def deleteMin(i: Int) : Option[Int]

  /** Union of heaps i and j. */
  def union(i: Int, j: Int) : Unit

  def shutDown: Unit
}

// ==================================================================
 
/** An object containing N concurrent binomial heaps, using locking for
  * union. */
class LockedMultiHeap(theLog: Log = NullLog, N: Int) extends MultiHeap{
  private val contents = Array.fill(N)(
    new ConcHeap(theLog = theLog, allowUnion = false))

  def shutDown = contents.foreach(_.shutdown)

  /** Offset used in the locking mechanism.  Must be more than the number of
    * threads.  FIXME */
  private val Offset = 1000 

  /** Objects controlling access to each heap.  A value k < Offset means that k
    * threads are using it, and it is not locked.  A value Offset + k means
    * that it is locked and k threads are using it. */
  private val locks = new java.util.concurrent.atomic.AtomicIntegerArray(N)

  /** Acquire access to heap i. */
  private def acquire(i: Int) = {
    var status = locks.get(i)
    while(status >= Offset || !locks.compareAndSet(i, status, status+1))
      status = locks.get(i)
  }

  /** Release use of heap i. */
  private def release(i : Int) = locks.getAndDecrement(i)

  /** Lock heap i; wait for all current threads to finish. */
  private def lock(i: Int) = {
    // Set flag to lock heap
    var status = locks.get(i)
    while(status >= Offset || !locks.compareAndSet(i, status, status+Offset))
      status = locks.get(i)
    // Wait until all current threads finished
    while(locks.get(i) > Offset){ }
  }

  private def unlock(i: Int) = {
    assert(locks.get(i) == Offset); locks.set(i, 0)
  }

  /** Insert key into heap i. */
  def insert(i: Int, key: Int) : Unit = {
    acquire(i); try{ contents(i).insert(key) } finally { release(i) }
  }

  /** Minimum key in heap i. */
  def minimum(i: Int) : Option[Int] = {
    acquire(i); try{ contents(i).minimum } finally { release(i) }
  }

  /** Delete minimum key in heap i. */
  def deleteMin(i: Int) : Option[Int] = {
    acquire(i); try{ contents(i).deleteMin } finally { release(i) }
  }

  /** Union of heaps i and j. */
  def union(i: Int, j: Int) : Unit = {
    assert(i != j)
    // Need to do acquire(i) and lock(j); do these in order of indices to
    // avoid deadlock.
    if(i < j){ acquire(i); lock(j) } else { lock(j); acquire(i) }
    try{ contents(i).union(contents(j)) }
    finally { release(i); unlock(j) }
  }
}

// ==================================================================

/** An object containing N concurrent binomial heaps, where union is done
  * without locking. */
class ConcurrentMultiHeap(
  theLog: Log = NullLog, N: Int, numMinsMin: Int = 1, numMinsDelMin: Int = 1)
    extends MultiHeap{

  private val contents = Array.fill(N)(
    new ConcHeap(theLog = theLog, allowUnion = true,
                 numMinsMin = numMinsMin, numMinsDelMin = numMinsDelMin))

  def shutDown = contents.foreach(_.shutdown)

  /** Insert key into heap i. */
  def insert(i: Int, key: Int) : Unit = contents(i).insert(key)

  /** Minimum key in heap i. */
  def minimum(i: Int) : Option[Int] = contents(i).minimum

  /** Delete minimum key in heap i. */
  def deleteMin(i: Int) : Option[Int] = contents(i).deleteMin

  /** Union of heaps i and j. */
  def union(i: Int, j: Int) : Unit = {
    assert(i != j); contents(i).union(contents(j)) 
  }
}
