package binomialHeap.tests

import ox.cads.experiments.{Experiments}

object NewPQueueExperiment{
  val Million = 1000000; val Billion = 1000*Million
  val Iters = 2*Million // # iterations by each thread
  val MaxKey = Billion // max key size
  var p = 64 // # threads
  var reps = 10 // # observations per experiment

  // Are we doing experiments making comparisons, resp, examining union?
  var comparison = false;  var union = false

  val cmd0 =
    "scala binomialHeap.tests.PQueueTimer --timing --iters "+Iters+
      " --maxKey "+MaxKey+" --reps 1"

  /* Each experiment is characterised by the following 5 parameters:
   * - iProb: the probability of doing an insertion
   * - dProb: the probability of doing a deleteMin
   * - mProb: the probability of doing a minimum
   * - uProb: the probability of doing a union (0.0 in comparisons)
   * - init: the initial size of the heap. 
   */

  /** Create a command. */
  def mkCmd(iProb: Double, dProb: Double, mProb: Double, uProb: Double,
            init: Int) =
    cmd0+" -p "+p+" --insertProb "+iProb+" --delMinProb "+dProb+
      " --minProb "+mProb+" --unionProb "+uProb+" --init "+init

  /** Do exactly Reps repetitions. */
  def statParams = new Experiments.Params(reps, reps, 0.05, 0.01)

  /** Convert time (in ns) for execution into number of operations per
    * second. */
  def rate(t: Long): Long = p.toLong*Iters*Billion/t

  /** A single measurement of cmd, giving time in nanos. */
  def measure(cmd: String) : Long = {
    val time = Experiments.timeProc(cmd, true)
    print(time/Million+" ") // in millis
    val theRate = rate(time)
    print(theRate+" ")
    theRate
  }

  val suffixes = List("", " --skiplist", " --LJskiplist")

  type ResultList = List[(Long,Long)]


  /** Run a particular experiment, comparing the different implementations. */
  def runInstance(iProb: Double, dProb: Double, mProb: Double, init: Int)
      : ResultList = {
    var results = List[(Long,Long)]()
    val cmd1 = mkCmd(iProb, dProb, mProb, 0.0, init)
    for(suf <- suffixes){
      val cmd = cmd1+suf; println(cmd)
      val (m, s) = Experiments.iterateMeasurement(measure(cmd), statParams)
      results = results ++ List((m.toLong, s.toLong))
      println; println((convert(m.toLong), convert(s.toLong))) // in ms
    }
    results
  }

  /** List of parameters to use in comparisons. */
  val params = Array(
    (0.5, 0.5, 0.0, 10000), (0.5, 0.5, 0.0, 100000), (0.5, 0.5, 0.0, Million),
    (0.4, 0.4, 0.2, 100000), (0.3, 0.3, 0.4, 100000),
    (0.6, 0.4, 0.0, 0), (0.7, 0.3, 0.0, 0)
  )

  // Strings for producing LaTeX
  val tab = " \t& "; val nl = " \\\\\n"

  def convert(n: Long) = (n+500)/1000+"K"

  def showResult(m: Long, s: Long) = convert(m)+"$\\pm$"+convert(s)

  /** Run the comparisons, producing results in a LaTeX tabular format. */
  def runComparisons = {
    val n = params.length
    val results = new Array[ResultList](n)
    for(i <- 0 until n){
      val (iProb, dProb, mProb, init) = params(i)
      results(i) = runInstance(iProb, dProb, mProb, init)
    }
    // Print results
    println
    for(i <- 0 until n){
      val (iProb, dProb, mProb, init) = params(i)
      print(iProb+tab+dProb+tab+mProb+tab+init+tab+
              results(i).map{case (m,s) => showResult(m,s)}.mkString(tab)+nl)
    }
  }

  val unionParams = Array(
    (0.38, 0.38, 0.19, 0.05, 100000, 4),
    (0.38, 0.38, 0.19, 0.05, 100000, 16),
    (0.4, 0.4, 0.19, 0.01, 100000, 16) // ,
    // (0.4, 0.4, 0.2, 0.0, 100000, 16)
  )


  /** Run experiments including union operation. */
  def runUnionExperiments = {
    val n = unionParams.length
    val results = new Array[(Long,Long)](n)
    for(i <- 0 until n){
      val (iProb, dProb, mProb, uProb, init, nh) = unionParams(i)
      val cmd =
        mkCmd(iProb, dProb, mProb, uProb, init)+" --allowUnion --numHeaps "+nh
      println(cmd)
      val (m, s) = Experiments.iterateMeasurement(measure(cmd), statParams)
      results(i) = (m.toLong, s.toLong) // in ms
      println; println((convert(m.toLong), convert(s.toLong))) // in ms
    }
    // Print results
    for(i <- 0 until n){
      val (iProb, dProb, mProb, uProb, init, nh) = unionParams(i)
      val (m,s) = results(i)
      print(iProb+tab+dProb+tab+mProb+tab+uProb+tab+init/1000+"K"+tab+nh+tab+
              showResult(m,s)+nl)
    }
  }

  def main(args: Array[String]) = {
    var i = 0
    while(i < args.length) args(i) match{
      case "-p" => p = args(i+1).toInt; i += 2
      case "--quick" => reps = 2; i += 1
      case "--union" => union = true; i += 1
      case "--comparison" => comparison = true; i += 1
      case a => println("Unrecognised argument: "+a); sys.exit
    }

    if(comparison) runComparisons
    if(union) runUnionExperiments
    if(!union && !comparison)
      println("Include either --union or --comparison flag")
  }

}
