package binomialHeap.concurrent

import java.util.concurrent.atomic._
import ox.cads.util.{ThreadID,Profiler}

/** The state of a node in a concurrent binary heap.
  * @param parent the parent of this node.
  * @param degree the degree of this node.
  * @param child the first child of this node.
  * @param next the next sibling of this node.
  * @param label an Label label, describing current updates happening. 
  */
class NodeState(
  val parent: Node = null, val degree: Int, val children: List[Node],
  val next: Node, val seq: Int, val label: Label = null){

  /** Is this unlabelled? */
  def unlabelled = label == null
  /** Does this have a null parent? */
  def parentless = parent == null
  /** Is this unlabelled and with a null parent? */
  def unlabelledParentless = unlabelled && parentless

  /** Create a new NodeState, cloning this but with next set to newNext. */
  def setNext(newNext: Node) =
    new NodeState(parent, degree, children, newNext, seq, label)

  /** Create a new NodeState, cloning this but with label set to newLabel. */
  def addLabel(newLabel: Label) =
    new NodeState(parent, degree, children, next, seq, newLabel)

  /** Create a new NodeState, cloning this but with parent set to null. */
  def clearParent =
    new NodeState(null, degree, children, next, seq, label)

  /** Is this a root node?  */
  def isRoot = parent == null || parent.deleted

  /** Get the next root node after this, skipping over nodes that are being
    * merged, and for which the t2 update has taken place.
    * @param thisNode the node for this state. */
  def getNextRoot(thisNode: Node) = label match{
    case MergeNext(t1, t2, state2) =>
      if(t2.getState.parent == t1) state2.next else t2
    case MergeParent(`thisNode`, _, _, t2, state2) =>
      if(t2.getState.parent == thisNode) state2.next else t2
    case _ => next
  }

  def labelledWith(ur: UnionReceiver): Boolean = label match{
    case hll: HeapLabelList => hll.getUR eq ur
    case _ => false
  }

  /** Create string representing first n nodes from here, to depth of d, where
    * this state corresponds to myNode with parent p.  Note this is
    * non-linearizable. 
    * Each node is represented by a string of the form
    * key[degree, label, parent]<children>
    * label is abbreviated to - or * if n=0
    * parent is given only if not as expected.
    */
  def toString(n: Int, d: Int = 5, myNode: Node, p: Node = null) : String =
    if(p == null) toStringRoot(n, d, myNode)
    else toStringChild(d, myNode, p)

  /** String giving the degree and label fields of a node. */
  private def showNode(n: Int) : String =
    degree+"," + (if(n == 0)( if(label == null) "-" else "*" ) else label)

  /** String representing children of a node. */
  private def showChildren(d: Int, myNode: Node) : String = 
    if(d>0){
      var st = "<"
      for(c <- children) st += c.toStringChild(d-1, myNode)
      st+">"
    }
    else if(degree==0) "<>" else "<..>"

  /** Create string representing first n root nodes from here, to depth of d,
    * where this state corresponds to root myNode.  Note this is
    * non-linearizable.
    * 
    * Each node is represented by a string of the form
    * key[degree, label, parent]<children>
    * label is abbreviated to - or * if n=0.
    * parent is given only if not as expected.
    */
  def toStringRoot(n: Int, d: Int, myNode: Node) : String = {
    def skip(t2: Node, state2: NodeState) = {
      assert(next == t2)
      "t2: "+next.toStringChild(d, null)+
        (if(state2.next != null) state2.next.toStringRoot(n-2, d) else "")
    }
    "["+showNode(n)+
      // parent info
      (if(parent == null) "R"
       else ", "+parent.key+(if(parent.deleted) "(deleted)" else "**")
      )+"]"+
      // children
      showChildren(d, myNode)+
      // continue
      (if(parent == null || parent.deleted){
         if(n > 0 && next != null) label match{
           case MergeNext(_, t2, state2) => skip(t2, state2)
           case MergeParent(`myNode`, _, _, t2, state2) => skip(t2, state2)
           case _ => next.toStringRoot(n-1, d)
         }
         else ""
       }
       else "^^"+parent.toString(n, d)
      )
  }

  /** Create string representing child node, to depth of d, where this state
    * corresponds to root myNode, and is expected to have parent p.  Note this
    * is non-linearizable.
    * 
    * Each node is represented by a string of the form
    * key[degree, label, parent]<children>
    * label is abbreviated to - or * if n=0.
    * parent is given only if not as expected.
    */
  def toStringChild(d: Int, myNode: Node, p: Node) : String = 
    "["+showNode(0)+
      // parent info
      (if(parent != p)(
         if(parent == null) ", null**" else ", "+parent.key+"**"
       )
       else ""
      )+"]"+
      // children
      showChildren(d, myNode)

  def compare(that: NodeState) : String = 
    (if(parent != that.parent) "Different parents " else "") +
      (if(degree != that.degree) "Different degrees " else "") +
      (if(children != that.children) "Different children " else "") +
      (if(next != that.next) "Different nexts " else "") +
      (if(! label.eq(that.label)) "Different label " else "") 

}

// ==================================================================

/** A node in a concurrent binomial heap. */
class Node(val key: Int){
  /** The state of this node, atomically updateable. */
  private val state = new AtomicReference[NodeState](
    new NodeState(
      parent = null, degree = 0, children = List(), next = null,
      seq = 0, label = null))

  /** Has the deletion of this node completed? */
  @volatile var deleted = false

  /** A heapIndirection pointing to this node's heap. */
  val heapIndirection = new AtomicReference[HeapIndirection](null)

  // def heapIndirectionCAS(oldHi: HeapIndirection, newHi: HeapIndirection) =
  //   oldHi == newHi || heapIndirection.compareAndSet(oldHi, newHi) ||
  //     heapIndirection.get == newHi

  /** Get the current state. */
  def getState : NodeState = state.get

  /** Get the current sequence number. */
  def getSeq : Int = state.get.seq

  /** Get the label. */
  def getLabel = getState.label

  /** Get the heap that this is currently part of. */
  def getHeap: ConcHeap = {
    // Perform path compression
    val hi = heapIndirection.get; val hi1 = hi.compress
    if(hi != hi1) heapIndirection.compareAndSet(hi, hi1)
    hi1.getHeap
  }
   //  heapIndirection.get.getHeap

  /** CAS the current state. */
  def compareAndSet(oldState: NodeState, newState: NodeState) =
    state.compareAndSet(oldState, newState)

  /** Set state to an initial state but with parent = p. */
  def setParentInitial(p: Node) = state.set(
    new NodeState(parent = p, degree = 0, children = List(),
                  next = null, seq = 0, label = null))

  /** Clear the state. */
  def clear = state.set(
    new NodeState(
      parent = null, degree = 0, children = List(), next = null,
      seq = 0, label = null))

  override def toString = toString(2, 1, null) // non-linearizable!

  private def keyString = key.toString+(if(deleted) "(deleted)" else "")

  /** Create string representing first n nodes from here, to depth of d. */
  def toString(n: Int, d: Int = 1, p: Node = null) : String =
    keyString + state.get.toString(n, d, this, p)

  /** Create string representing first n nodes from here, to depth of d,
    * assuming this is a child node. */
  def toStringChild(d: Int, p: Node) =
    keyString + state.get.toStringChild(d, this, p)

  /** Create string representing first n nodes from here, to depth of d,
    * assuming this is a child node. */
  def toStringRoot(n: Int, d: Int) =
    keyString + state.get.toStringRoot(n, d, this)

  /** Get the state of this node, clearing the parent if it points to a fully
    * deleted node. */
  def maybeClearParent() : NodeState = {
    val oldState = state.get; val p = oldState.parent
    if(p != null && p.deleted){
      val newState = oldState.clearParent
      if(state.get == oldState && compareAndSet(oldState, newState)) newState
      else maybeClearParent() // retry: this might have been merged again
    }
    else oldState
  }

}

// ==================================================================
