package binomialHeap.tests

import ox.cads.experiments.{Experiments,Graphs}
import java.io._

/** A program to carry out various experiments on the binomial heap.
  * 
  * Command line arguments:
  * -p number of threads
  * --iters number of iterations per run
  * --reps number of repetitions (when profiling)
  * --insertProb, --minProb, --delMinProb, --unionProb probability of an
  * operation being an insert, minimum, deleteMin or union.
  * --maxKey maximum key to use
  * --init number of items in heap initially
  * --DES do discrete event simulation
  * --quick do quick run, with looser requirements on confidence intervals
  */
object PQueueExperiment{
  var iters = 500000 // # iterations per run
  var reps = 1 // # runs per measurement
  var p = 8 // 32 for Gimli
  var insertProb = 0.5 // proportion of insertions
  var minProb = 0.2 // proportion of minimum operations
  var delMinProb = 0.2 // proportion of deleteMins
  var unionProb = 0.1
  var init = 0 // # values in heap initially
  var maxKey = 1000000

  // Tuning paramters.  Below, commented out parameters are (roughly) optimal
  // for the old deletion algorithm.

  var tidyRatio = 2 // inverse frequency of tidying
  var recTidy = true // are we recursively tidying?
  var A = 16; var B = 16
  // var numMinsMin = 4 // number of min values to record in minimum
  var numMinsMin = 2 // number of min values to record in minimum
  // var numMinsDelMin = 6 // number of min values to record in delMin
  var numMinsDelMin = 16 // number of min values to record in delMin
  var retrieving = false // are we using the backtracking strategy?
  // var backtracking = false // are we using the backtracking strategy?
  var backtracking = false // are we using the backtracking strategy?
  var useSkiplist = false // are we using the skiplist implementation?
  var useLJSkiplist = false // are we using the skiplist implementation?
  var doDES = false // are we doing a discrete event simulation
  var doTimeDependent = false // are we doingthe time-dependent use case?

  // Command to invoke PQueueTimer
  def cmd0 =
    "scala binomialHeap.tests.PQueueTimer --timing --reps 1 -p "+p+
      " --iters "+iters+" --maxKey "+maxKey+" --init "+init+
      (if(doDES) " --DES"
       else if(doTimeDependent)
         " --timeDependent --minProb "+minProb
       else
         " --insertProb "+insertProb+" --minProb "+minProb+" --delMinProb "+
         delMinProb
      )+
      (if(useSkiplist) " --skiplist"
       else if(useLJSkiplist) " --LJskiplist"
       else(
         " --tidyRatio "+tidyRatio+
           (if(recTidy) " --recTidy" else "")+
           " --A "+A+" --B "+B+
           (if(backtracking) " --backtracking" else "")+
           (if(retrieving) " --retrieving" else "")+
           " --numMinsMin "+numMinsMin+" --numMinsDelMin "+numMinsDelMin
       )
      )

  val Billion = 1000000000
  val Million = 1000000

  /** Parameters defining statistical analysis of experiments: do at least 5
    * samples; repeat until 95% confidence half-interval is at most 1% of mean
    * or until 50 repetitions done. */
  val normalParams = new Experiments.Params(10, 50, 0.05, 0.01) 
  val quickParams = new Experiments.Params(5, 0.05, 0.1) // looser params
  val strictParams = new Experiments.Params(40, 80, 0.05, 0.01)
  val veryStrictParams = new Experiments.Params(100, 200, 0.05, 0.01)
  var params = normalParams

  /** A single measurement of cmd, giving time in nanos. */
  def measure(cmd: String) : Double = {
    val time = Experiments.timeProc(cmd, true)
    print(time/Million+" ") // in millis
    time.toDouble
  }

  /** Standard options to be passed to pgfplots. */
  val stdOpts =
    Array("ylabel = Time (s)", "xtick = data","legend pos = north east",
          "height = 0.6\\textheight", "width = 0.9\\textwidth")

  /** LaTeX newline. */
  def nl = " \\\\\n"

  /* Strings describing some parameters. */
  def paramsString0 =
    (if(doDES) "Discrete event simulation." else "")+
    "iters = "+iters+"; p = "+p+"; init = "+init
  def tidyParamsString = "tidyratio = "+tidyRatio+" recTidy = "+recTidy
  def numMinsString =
    "numMinsMin = "+numMinsMin+"; numMinsDelMin = "+numMinsDelMin
  def paramsString =
    paramsString0+nl+
      (if(!doDES)
         "insertProb = "+insertProb+
         "; minProb = "+minProb+"; delMinProb = "+delMinProb
       else "")

  /** Output graphString, both to standard out and to file fname. */
  def outputGraph(graphString: String, fname: String) = {
    println(graphString)
    val fname1 =
      if(useSkiplist) "skiplist-"+fname
      else if(useLJSkiplist) "LJskiplist-"+fname else fname
    Graphs.writeStandAloneFile(fname1, graphString)
    println("Output written to "+fname1)
  }

  // =======================================================

  /** Experiment to investigate effect of tidying probability and recursive
    * tidying. */
  def tidyProbExperiment() = {
    val ratios =  (0 to 5).map(1 << _).toArray
    // Array to store results, indexed by whether recursive tidying is done (0
    // = false; 1 = true), and index into ratios.
    val results = Array.ofDim[(Double, Double)](2,ratios.length)

    // Perform experiment
    for(j <- 0 until 2; i <- 0 until ratios.length){
      val recTidy = (j==1) // Do we tidy recursively?
      val recTidyString = if(recTidy) " --recTidy" else ""
      val tidyRatio = ratios(i);
      println("tidyRatio = "+tidyRatio+"; recTidy = "+recTidy)
      val cmd = cmd0+" --tidyRatio "+tidyRatio+recTidyString
      println(cmd)
      // Iterate the measurement
      val (m, s) = Experiments.iterateMeasurement(measure(cmd), params)
      println; println(m/1000000,s/1000000); println // in millis
      results(j)(i) = (m/Billion,s/Billion) // in seconds
    }

    // Produce graphs
    val title =
      "{\\begin{tabular}{l}\n"+
        "Effect of tidyRatio and recursive tidying. \\\\\n"+paramsString+
        "numMinsMin = "+numMinsMin+"; numMinsDelMin = "+numMinsDelMin+
        "\n\\end{tabular}}"
    val options = 
      Array[String]("title = "+title, "xlabel = Ratio", "log basis x = 2") ++
        stdOpts
    val graphString =
      Graphs.makeLogXGraph(options, Array("false", "true"), ratios, results)
    outputGraph(graphString, "tidyProbExperiment.tex")
  }

  // =======================================================

  /** Experiment to estimate number of minimums that should be stored during
    * traversing.  
    * Needs to be done with --strict flag. */
  def numMinsExperiment = {
    // Number of minimums to consider
    val minsMin = (0 to 3).map(1 << _).toArray
    val minsDelMin = (0 to 6).map(1 << _).toArray
    // array to hold results, indexed by indices of minsMin, minsDelMin
    val results =
      Array.ofDim[(Double, Double)](minsMin.length, minsDelMin.length)

    // Perform experiment
    for(i <- 0 until minsMin.length; j <- 0 until minsDelMin.length){
      val numMinsMin = minsMin(i); val numMinsDelMin = minsDelMin(j)
      println("numMinsMin = "+numMinsMin+"; numMinsDelMin = "+numMinsDelMin)
      val cmd =
        cmd0+" --numMinsMin "+numMinsMin+" --numMinsDelMin "+numMinsDelMin
      println(cmd)
      // Iterate the measurement
      val (m, s) = Experiments.iterateMeasurement(measure(cmd), params)
      println; println(m/1000000,s/1000000); println // in millis
      results(i)(j) = (m/Billion,s/Billion) // in seconds
    }

   // Produce graphs
    val title =
      "{\\begin{tabular}{l}\n"+
        "Effect of varying the number of minima stored while traversing. \\\\\n"+
        paramsString+";\\\\\n tidyRatio = "+tidyRatio+
        "; recTidy = "+recTidy+"\n\\end{tabular}}"
    val options = 
      Array[String](
        "title = "+title, 
        "xlabel = Number of minima recorded while traversing in deleteMin",
        "log basis x = 2"
      ) ++ stdOpts
    val graphString = Graphs.makeLogXGraph(
      options, minsMin.map(_.toString), minsDelMin, results)
    outputGraph(graphString, "numMinsExperiment.tex")
  }

  // ==================================================================

  /** Experiment investigating the effect of varying the proportion of different
    * operations. */
  def paramsExperiment = {
    // To avoid rounding errors, each probability will be stored as an integer
    // n with 0 <= n <= 10, and interpreted as the probability n/10.
    def prob(n: Int) : Double = n/10.0
    val minProbs = Array(0,2,4) // Array(1, 2, 3, 4)
    val insertProbs = (0 to 10).map(prob(_)).toArray
    val results =
      Array.ofDim[(Double, Double)](minProbs.length, insertProbs.length)

    for(i <- 0 until minProbs.length){
      val minn = minProbs(i); minProb = prob(minn)
      for(j <- 1 to 10-minn){
        val insertProb = prob(j)
        val delMinProb = 1.0-minProb-insertProb
        val cmd = cmd0+" --minProb "+minProb+" --delMinProb "+delMinProb+
          " --insertProb "+insertProb
        println(cmd)
        val (m, s) = Experiments.iterateMeasurement(measure(cmd), params)
        println; println(m/1000000,s/1000000); println // in millis
        results(i)(j) = (m/Billion,s/Billion) // in seconds
      }
    }

    // Produce graphs
    val title =
      "{\\begin{tabular}{l}\n"+
        "Effect of varying proportions of operations."+nl+
        paramsString0+nl+tidyParamsString+nl+numMinsString+
        "\n\\end{tabular}}"
    val options = 
      Array[String]("title = "+title, "xlabel = probability of insertion") ++
        stdOpts
    val graphString = Graphs.makeLinearGraph(
        options, minProbs.map(_.toString), insertProbs.map(_.toString), results)
    outputGraph(graphString, "paramsExperiment.tex")
  }

  // ==================================================================

  def scalingExperiment = {
    val ps = (4 to 64 by 4).toArray
    val results = Array.ofDim[(Double, Double)](1, ps.length)
    // Throughput, in ops per second, given argument of time in nanos, and
    // number p of threads.
    def mkThroughput(t: Double, p: Int) = p.toDouble*iters/(t/Billion)

    for(i <- 0 until ps.length){
      val p = ps(i); val cmd = cmd0+" -p "+p
      println(cmd)
      val (m, s) =
        Experiments.iterateMeasurement(mkThroughput(measure(cmd), p), params)
      println; println(m, s) // in ops/sec
      results(0)(i) = (m/Million, s/Million) // in M ops/sec
    }

    // Produce graphs
    val title =
      "{\\begin{tabular}{l}\n"+
        "Effect of scaling."+nl+paramsString+(
          if(!useSkiplist && !useLJSkiplist)
            nl+tidyParamsString+nl+numMinsString
          else ""
        )+
        "\n\\end{tabular}}"
    val options = Array[String](
      "title = "+title, "xlabel = number of threads",
      "ylabel = M operations/sec") ++
      stdOpts.tail
    val graphString = Graphs.makeLinearGraph(
        options, Array("XXX"), ps.map(_.toString), results)
    outputGraph(graphString, "scalingExperiment.tex")
  }

  // ==================================================================

  /** Experiment to tune the boundOffSet of the LJPQueue */
  def LJPQueueBoundExperiment = {
    val bounds = (4 to 84 by 8).toArray
    val results = Array.ofDim[(Double, Double)](1, bounds.length)

    for(i <- 0 until bounds.length){
      val bound = bounds(i); val cmd = cmd0+" --boundOffset "+bound
      println(cmd)
      val (m, s) =
        Experiments.iterateMeasurement(measure(cmd), params)
      println; println(m/1000000,s/1000000) // in millis
      results(0)(i) = (m/Billion,s/Billion) // in seconds
    }
    // Produce graphs
    val title =
      "{\\begin{tabular}{l}\n"+
        "Experiment to tune boundOffset for LJ Priority Queue."+nl+
        paramsString+
        "\n\\end{tabular}}"
    val options = Array[String](
      "title = "+title, "xlabel = boundOffset",
      "ylabel = time (ms)") ++
      stdOpts.tail
    val graphString = Graphs.makeLinearGraph(
        options, Array("XXX"), bounds.map(_.toString), results)
    outputGraph(graphString, "boundOffsetExperiment.tex")
  }


  // =======================================================

  def profile = {
    val rt = java.lang.Runtime.getRuntime()
    val cmd =
      cmd0+" --profile"
      println(cmd)
    for(i <- 0 until reps){
      val proc = rt.exec(cmd)
      val isr = new InputStreamReader(proc.getInputStream, "ISO-8859-1")
      val fromProc = new BufferedReader(isr)
      var t = -1L // last value that looked like an integer
      var tstr = fromProc.readLine
      while(tstr != null){
        if(tstr.nonEmpty && tstr.forall(_.isDigit)) t = tstr.toLong
        print("*"+tstr+"; ")
        tstr = fromProc.readLine
      }
      // assert(!flushErr(proc)) // check for errors
      proc.waitFor
      println
    }
  }

  // ==================================================================

  /** Experiment to examine the effects of the "backtracking" and "retrieving"
    * tactics in deleteMin. */
  def delMinParamsExperiment = {
    for(backtracking1 <- List(true, false); retrieving1 <- List(true, false)){
      backtracking = backtracking1; retrieving = retrieving1
      println(cmd0)
      val (m, s) = Experiments.iterateMeasurement(measure(cmd0), params)
      println; println(m/1000000,s/1000000); println // in millis
    }
  }

  // ==================================================================

  /** Experiment varying the A and B paremeters. */ 
  def ABExperiment = {
    val As = Array(4, 8, 12, 16, 20, 24, 28)
    val Bs = Array(8, 12, 16, 20, 24)
    val results = Array.ofDim[(Double, Double)](Bs.length, As.length)

    for(i <- 0 until Bs.length){
      val B = Bs(i)
      for(j <- 0 until As.length){
        val A = As(j); val cmd = cmd0+" --A "+A+" --B "+B; println(cmd)
        val (m, s) = Experiments.iterateMeasurement(measure(cmd), params)
        println; println(m/1000000,s/1000000); println // in millis
        results(i)(j) = (m/Billion,s/Billion) // in seconds
      }
    }

    // Produce graphs
    val title =
      "{\\begin{tabular}{l}\n"+
        "Effect of varying parameters A and B."+nl+
        paramsString0+nl+tidyParamsString+nl+numMinsString+
        "\n\\end{tabular}}"
    val options = Array[String]("title = "+title, "xlabel = A") ++ stdOpts
    val graphString = Graphs.makeLinearGraph(
      options, plotNames = Bs.map(_.toString),
      xLabels = As.map(_.toString), results)
    outputGraph(graphString, "ABExperiment.tex")
  }

  // ==================================================================

  /** Basic timing run. */
  def basic = {
    println(cmd0)
    val (m, s) = Experiments.iterateMeasurement(measure(cmd0), params)
    println; println(m/1000000,s/1000000); println // in millis
  }

  // =======================================================

  def main(args: Array[String]) = {
    // Parse arguments. 
    var i = 0
    var doTidyProbExperiment = false; var doNumMinsExperiment = false
    var doProfile = false; var doDelMinParamsExperiment = false
    var doBasic = false; var doParamsExperiment = false
    var doScalingExperiment = false; var doLJPBoundsExperiment = false
    var doABExperiment = false
    while(i < args.length) args(i) match{
      case "-p" => p = args(i+1).toInt; i += 2
      case "--iters" => iters = args(i+1).toInt; i += 2
      case "--reps" => reps = args(i+1).toInt; i += 2
      case "--maxKey" => maxKey = args(i+1).toInt; i += 2
      case "--insertProb" => insertProb = args(i+1).toDouble; i += 2
      case "--minProb" => minProb = args(i+1).toDouble; i += 2
      case "--delMinProb" => delMinProb = args(i+1).toDouble; i += 2
      case "--init" => init = args(i+1).toInt; i += 2
      case "--DES" => doDES = true; i += 1
      case "--timeDependent" => doTimeDependent = true; i += 1
      // Parameters for statistical analysis
      case "--quick" => params = quickParams; i += 1
      case "--strict" => params = strictParams; i += 1
      case "--veryStrict" => params = veryStrictParams; i += 1
      // Parameters for ConcHeap
      case "--tidyRatio" => tidyRatio = args(i+1).toInt; i += 2
      case "--recTidy" => recTidy = true; i += 1
      case "--A" => A = args(i+1).toInt; i += 2
      case "--B" => B = args(i+1).toInt; i += 2
      case "--backtracking" => backtracking = true; i += 1
      case "--noBacktracking" => backtracking = false; i += 1
      case "--retrieving" => retrieving = true; i += 1
      case "--numMinsDelMin" => numMinsDelMin = args(i+1).toInt; i += 2
      // Which heap to use
      case "--skiplist" => useSkiplist = true; i += 1
      case "--LJskiplist" => useLJSkiplist = true; i += 1
      // Which experiments to do?
      case "--tidyProb" => doTidyProbExperiment = true; i += 1
      case "--numMins" => doNumMinsExperiment = true; i += 1
      case "--delMinParams" => doDelMinParamsExperiment = true; i += 1
      case "--params" => doParamsExperiment = true; i += 1
      case "--scaling" => doScalingExperiment = true; i += 1
      case "--profile" => doProfile = true; i += 1
      case "--AB" => doABExperiment = true; i += 1
      case "--basic" => doBasic = true; i += 1
      case "--LJPQueueBounds" => doLJPBoundsExperiment = true; i += 1
      case _ => println("Unrecognised argument: "+args(i)); sys.exit
    }

    // run experiments
    if(doNumMinsExperiment) numMinsExperiment
    if(doTidyProbExperiment) tidyProbExperiment
    if(doDelMinParamsExperiment) delMinParamsExperiment
    if(doParamsExperiment) paramsExperiment
    if(doScalingExperiment) scalingExperiment
    if(doLJPBoundsExperiment) LJPQueueBoundExperiment
    if(doProfile) profile
    if(doABExperiment) ABExperiment
    if(doBasic) basic
  }
}
