package binomialHeap.tests

import ox.cads.util.{Profiler,ThreadUtil,ThreadID,SamplingProfiler}
import scala.util.Random

/** A timer for a priority queue. 
  * 
  * Flags affecting type of run:
  * --DES do the discrete event simulation case
  * --skiplist use the basic skiplist implementation
  * --LJskiplist use the Linden-Jonsson skiplist
  * 
  * Flags affecting details of each run:
  * -p number of threads
  * --iters number of iterations per run
  * --reps number of runs
  * --insertProb, --minProb, --delMinProb probability of an operation being an 
  * insert, minimum of deleteMin
  * --maxKey maximum key to use
  * --init number of values in heap initially
  * 
  * Flags affecting how run is treated:
  * --timing perform formal timing, giving result in ns
  * --useLog perform logging
  * --profile perform profiling
  * --maxInsertTime tThreshold for giving profiling info about inserts (in ms)
  * --maxDelMinTime tThreshold for giving profiling info about delMins (in ms)
  * 
  * Tuning parameters:
  * --numMinsMin, --numMinsDelMin number of min keys recorded during traversal
  * --tidyRatio ratio of insert/delete operations to tidying
  * --recTidy should tidy act recursively?
  * --A --B the A and B parameters for tidying
  * --backtracking use the backtracking tactic
  * --retrieving use the retrieving tactic
  * --boundOffset the bound offset to use with the LJ skiplist.
  */
object PQueueTimer{
  var iters = 1000 // 0  // # iterations in each run
  var reps = 1000  // # repetitions
  var p = 8 // #threads
  var MaxKey = 1000000
  var numHeaps = 4 // number of heaps
  // Probabilities of operations
  var insertProb = 0.4; var minProb = 0.25; var delMinProb = 0.25
  var unionProb = 0.1
  assert(insertProb+minProb+delMinProb+unionProb == 1.0)

  val Million = 1000000; val Billion = 1000*Million
  var profiling = false // Perform profiling?
  // Threshold for giving profiling info about inserts and delMins (in ms).
  var maxInsertTime = 12000; var maxDelMinTime = 25000
  var maxTime = 35000

  /** A worker. */
  def worker(me: Int, heaps: Array[binomialHeap.PriorityQueue]) = {
    val random = new Random(me+Random.nextInt)
    for(i <- 0 until iters){
      val h = random.nextInt(numHeaps) // heap to use
      val rand = random.nextFloat
      if(rand <= insertProb){ // insert
        if(false && profiling)
          Profiler.time("insert"){ heaps(h).insert(random.nextInt(MaxKey)) }
        else heaps(h).insert(random.nextInt(MaxKey))
      }
      else if(rand <= insertProb+minProb){ // minimum
        if(false && profiling) Profiler.time("minimum"){ heaps(h).minimum }
        else heaps(h).minimum
      }
      else if(rand <= insertProb+minProb+delMinProb){ // delMin
        if(false && profiling) Profiler.time("deleteMin"){ heaps(h).deleteMin }
        else heaps(h).deleteMin
      } 
      else{ // union
        val g0 = random.nextInt(numHeaps-1); val g = if(g0 >= h) g0+1 else g0
        assert(g != h)
        val giver = heaps(g).asInstanceOf[binomialHeap.concurrent.Union]
        heaps(h).union(giver)
      }
      //  if(rand < 0.1) heap.minimum
      // else if(rand < 0.3 + 0.5*i/iters) heap.deleteMin
      // else heap.insert(random.nextInt(MaxKey))
      // Prob of insert is initially 0.7, but falls to 0.2.  Prob of delete is
      // initially 0.2, but rises to 0.7.
    }
  }

  /** A worker whose behaviour is time dependent.  Each operation is a minimum
    * with probability minProb.  The probability of an insertion decreases
    * linearly from 1-minProb to 0; and the probability of a deleteMin
    * increases linearly from 0 to 1-minProb. */
  def timeDependentWorker(me: Int, heap: binomialHeap.PriorityQueue) = {
    val random = new Random(me+Random.nextInt)
    for(i <- 0 until iters){
      if(random.nextFloat <= minProb) heap.minimum
      else if(random.nextInt(iters) < i) heap.deleteMin
      else heap.insert(random.nextInt(MaxKey))
    }
  }

  /** A worker in a discrete event simulation. */
  def DESWorker(me: Int, heap: binomialHeap.PriorityQueue) = {
    val random = new Random(me+Random.nextInt)
    var delMinTime = 0L; var insertTime = 0L
    if(profiling) for(i <- 0 until iters){ // perform timing
      val t0 = java.lang.System.nanoTime
      val res = heap.deleteMin
      val t1 = java.lang.System.nanoTime
      delMinTime += (t1-t0)
      if(t1-t0 > 800*Million) // deleteMin time > 800ms
        println(ThreadID.get+":deleteMin:"+res+":"+(t1-t0)/Million)
      res match{
        case Some(x) =>
          val y = x+1000+random.nextInt(1000)
          heap.insert(y)
          val t2 = java.lang.System.nanoTime
          insertTime += (t2-t1)
          if(t2-t1 > 300*Million) // insert time > 300ms
            println(ThreadID.get+":insert("+y+"):"+(t2-t1)/Million)
        case None => {}
      }
    } // end of timing branch
    else for(i <- 0 until iters){
      heap.deleteMin match{
        case Some(x) => heap.insert(x+1000+random.nextInt(1000))
        case None => {}
      }
    } // end of else-for
    // if(insertTime/Million > maxInsertTime ||
    //      delMinTime/Million > maxDelMinTime) println(
    //   ThreadID.get+": inserts:"+insertTime/Million+
    //     "; delMins:"+delMinTime/Million+"\n"+
    //     (if(insertTime/Million > maxInsertTime)
    //        "end CAS fail "+Profiler.getCounter("end CAS fail"+ThreadID.get)+
    //        "; below CAS fail "+Profiler.getCounter("below CAS fail"+ThreadID.get)+
    //        "; helpDelete "+Profiler.getCounter("helpDelete"+ThreadID.get)+
    //        "; helpMerge "+Profiler.getCounter("helpMerge"+ThreadID.get)+
    //        "; followParent "+Profiler.getCounter("follow parent"+ThreadID.get)+
    //        "\n"
    //      else ""
    //     )+
    //     (if(delMinTime/Million > maxDelMinTime)
    //        "delMin restart "+Profiler.getCounter("delMin restart"+ThreadID.get)+
    //        "; tryDelete "+Profiler.getCounter("tryDelete"+ThreadID.get)
    //      else "")
    // ) // end of println
  }

  /** Main function. */
  def main(args: Array[String]) = {
    // Parse arguments. 
    var i = 0; var logRoots = 0; var timing = false;
    var useSkiplist = false; var useLJSkiplist = false
    var numMinsMin = 4; var numMinsDelMin = 6; var useLog = false
    var tidyRatio = 64; var recTidy = false; var allowUnion = false
    var A = 16; var B = 16
    var backtracking = false; var retrieving = false
    var boundOffset = 20; var init = 0;
    var doDES = false; var doTimeDependent = false
    var verbose = false
    while(i < args.length) args(i) match{
      // Type of run
      case "--DES" => doDES = true; i += 1
      case "--timeDependent" => doTimeDependent = true; i += 1
      case "--skiplist" => useSkiplist = true; i += 1 
      case "--LJskiplist" => useLJSkiplist = true; i += 1 
      // Details of run
      case "-p" => p = args(i+1).toInt; i += 2
      case "--iters" => iters = args(i+1).toInt; i += 2 
      case "--reps" =>  reps = args(i+1).toInt; i += 2
      case "--maxKey" => MaxKey = args(i+1).toInt; i += 2
      case "--insertProb" => insertProb = args(i+1).toFloat; i += 2
      case "--minProb" => minProb = args(i+1).toFloat; i += 2
      case "--delMinProb" => delMinProb = args(i+1).toFloat; i += 2
      case "--unionProb" => unionProb = args(i+1).toFloat; i += 2
      case "--numHeaps" => numHeaps = args(i+1).toInt; i += 2
      case "--init" => init = args(i+1).toInt; i += 2
      case "--allowUnion" => allowUnion = true; i += 1
      // How run treated
      case "--timing" => timing = true; i += 1 
      case "--profile" => profiling = true; i += 1
      case "--useLog" => useLog = true; i += 1
      case "--maxInsertTime" => maxInsertTime = args(i+1).toInt; i += 2
      case "--maxDelMinTime" => maxDelMinTime = args(i+1).toInt; i += 2
      case "--maxTime" => maxTime = args(i+1).toInt; i += 2
      case "--verbose" => verbose = true; i += 1
      // Parameters to tune
      case "--numMinsMin" => numMinsMin = args(i+1).toInt; i += 2
      case "--numMinsDelMin" => numMinsDelMin = args(i+1).toInt; i += 2
      case "--tidyRatio" => tidyRatio = args(i+1).toInt; i += 2
      case "--recTidy" => recTidy = true; i += 1
      case "--backtracking" => backtracking = true; i += 1
      case "--A" => A = args(i+1).toInt; i += 2
      case "--B" => B = args(i+1).toInt; i += 2
      case "--retrieving" => retrieving = true; i += 1
      case "--boundOffset" =>
        assert(useLJSkiplist); boundOffset = args(i+1).toInt; i += 2
      case _ => println("Unrecognised argument: "+args(i)); sys.exit 
    }

    if(doDES) assert(init >= p)
    else if(!doTimeDependent){
      assert(Math.abs(insertProb+minProb+delMinProb+unionProb-1.0) < 0.0000001,
             "probabilities sum to "+(insertProb+minProb+delMinProb+unionProb))
      if(numHeaps == 1)
        assert(unionProb == 0.0, "single heap but union allowed")
      if(unionProb > 0.0) assert(allowUnion, "need --allowUnion")
    }
    Profiler.setWorkers(p)
    val dotStep =
      if(useLog) 1 max (20000/(p*iters)) else 1 max (200000/(p*iters))

    // Now run the tests
    //val random = new scala.util.Random(scala.util.Random.nextInt)
    // Profiler.clear
    var time = 0L

    for(i <- 0 until reps){ 
      val theLog = 
        if(useLog) ox.cads.util.ScreenLog
        // if(useLog) new ox.cads.util.TSLog("/tmp/log", p)
        else ox.cads.util.NullLog
      // Function to make a heap
      def mkHeap =
        if(useSkiplist) new binomialHeap.skiplist.SLHeap("")
        else if(useLJSkiplist) new binomialHeap.skiplist.LJPQueue(boundOffset)
        else new binomialHeap.concurrent.ConcHeap(
          theLog = theLog, allowUnion = allowUnion,
          MaxRetries = 2000 max (p*200),
          numMinsMin = numMinsMin, numMinsDelMin = numMinsDelMin,
          tidyRatio = tidyRatio, RecTidy = recTidy, A = A, B = B,
          backtracking = backtracking, retrieving = retrieving)
      // Create the heaps, and seed them with init values
      val heaps = Array.fill(numHeaps)(mkHeap)
      for(h <- 0 until numHeaps; _ <- 0 until init)
        heaps(h).insert(Random.nextInt(MaxKey))
      // Function to create the system
      def system = 
        if(doDES) ThreadUtil.runIndexedSystem(p, DESWorker(_, heaps(0)))
        else if(doTimeDependent)
          ThreadUtil.runIndexedSystem(p, timeDependentWorker(_, heaps(0)))
        else ThreadUtil.runIndexedSystemStrict(p, worker(_, heaps))

      // Start timing
      val t0 = java.lang.System.nanoTime
      if(profiling){
        // Create the profiler
        def filter(f: java.lang.StackTraceElement) =
          f.getClassName.contains("NanoSpin") ||
            (SamplingProfiler.notAPIFrame(f) &&
               !f.getClassName.startsWith("binomialHeap.tests"))
        val printer = SamplingProfiler.print(filter = filter)(_)
        val profiler = new SamplingProfiler(interval = 100, print = printer)
        // Run the system
        val profile = profiler(system, false)
        val duration = java.lang.System.nanoTime-t0; time += duration
        val durationMs = duration / Million
        if(durationMs > maxTime || verbose || (reps > 1 && i%10 == 0)){
          if(durationMs > maxTime) println("\n*** "+durationMs+" *** ")
          else println(durationMs)
          println(profile); Profiler.report
        }
        else print(durationMs)
        Profiler.clear
      }
      else{
        try{ system } finally { theLog.shutdown }
        val duration = java.lang.System.nanoTime-t0; time += duration
      }

      if(reps>1 && !timing && i%dotStep == 0) print(".")
      if(reps>1 && !timing && i%(1000*dotStep) == 0) println("\n"+i)
    } // end of for loop

    if(timing) println(time)
    else if(!profiling || reps>1) println((time)/Million+" ms")
    if(!timing && profiling){ println; Profiler.report }
  }
}
