package binomialHeap.skiplist

import java.util.concurrent.atomic._
import ox.cads.atomic.AtomicPair
import ox.cads.util.{ThreadID,Profiler}

/** A concurrent priority queue implemented using a skiplist.  Based on "Fast
  * and Lock-Free Concurrent Priority Queues for Multi-Thread Systems. */

class SLNode(val key: Long, val level: Int){
  /** The level at which this node is correctly inserted. */
  @volatile var validLevel = -1

  /** Next pointers at each level, each paired with a boolean. */
  val next = Array.fill(level)(new AtomicPair[SLNode, Boolean](null, false))
  // Each boolean gets set to true to indicate that this node is being
  // deleted.  The pair gets set to (null, true) to indicate that the node is
  // fully deleted.

  /** Prev pointer, at the bottom level. */
  @volatile var prev : SLNode = null

  /** Is this node deleted? */
  val deleted = new AtomicBoolean(false)

  override def toString = (key >> 32).toString+(if(deleted.get) "X" else "")

  def toString(n: Int = 0) : String = {
    val nodeString = (key >> 32).toString+(if(deleted.get) "X" else "")
    val (nextNode,_) = next(0).get
    val nextSt =
      if(nextNode == null) "" else if(n == 0) "..." else nextNode.toString(n-1);
    (key >> 32).toString+" "+nextSt
  }
}

// ==================================================================

class SLHeap(logName: String) extends binomialHeap.PriorityQueue{
  /** Maximum number of levels in any node. */
  private val MaxLevel = 32

  private val TopLevel = MaxLevel-1 // top valid level

  // Logging and assertions
  val theLog =
    if(logName != "") new ox.cads.util.ArrayBufferLog(logName)
    else ox.cads.util.NullLog

  private def log(st: => String) = theLog(ThreadID.get+": "+st)

  def shutdown = theLog.shutdown

  private def myAssert(ok: Boolean, st: => String) : Unit =
    theLog.assert(ok, ThreadID.get+": "+st+"\n"+this)

  private def myAssert(ok: Boolean) : Unit = myAssert(ok: Boolean, "")

  /** Create the key used internally for ordering in the skip list.  Append 32
    * random bits to key to ensure uniqueness. */
  private def mkLongKey(key0: Int) = {
    val key : Long = (key0.toLong << 32) | (scala.util.Random.nextLong >>> 32)
    assert(extractIntKey(key) == key0)
    key
  }

  /** Extract Int key from Long key, removing random bits. */
  private def extractIntKey(key: Long) = (key >> 32).toInt

  /** Sentinel head node. */
  private val head = new SLNode(Long.MinValue, MaxLevel)
  /** Sentinel tail node. */
  private val tail = new SLNode(Long.MaxValue, MaxLevel)

  // Initialise head's next pointers
  for(l <- 0 until MaxLevel) head.next(l).set((tail, false))
  head.validLevel = MaxLevel; tail.validLevel = MaxLevel

  /** String to represent the state of the queue. */
  override def toString = "List: "+head.toString(20)

  // private def myAssert(ok: Boolean, st: => String) : Unit =
  //   if(!ok){
  //     val st1 = st; val thisSt = this.toString
  //     if(logName != ""){
  //       log("***ERROR*** "+st1+"\n"+this); theLog.shutdown
  //     }
  //     else{ // write to file fname
  //       val fname = "/tmp/error"+ThreadID.get
  //       val thisLog = new ox.cads.util.ArrayBufferLog(fname)
  //       thisLog(st1+"\n"+thisSt); thisLog.shutdown
  //       println("Error written to "+fname)
  //     }
  //     sys.error(st1+"\n"); sys.exit()
  //   }

  // private def myAssert(ok: Boolean) : Unit = myAssert(ok, "")

  private def readNode(pair: (SLNode, Boolean)) : SLNode =
    pair match { case (n, mark) => if(mark) null else n }

  /** Get the next node after node at this level.  Help with deletes, retreating
    * if necessary.  The normal usage is
    * val (n, next) = readNext(node, level); node = n
    * @return a pair (n, next) such that n is the new value to be used for
    * node, and next is the next node. */
  private def readNext(node: SLNode, level: Int) : (SLNode, SLNode) = {
    // log("readNext "+node)
    myAssert(node != tail && node != null, node.toString)
    var node1 = // current node
      if(node.deleted.get) helpDelete(node, level) else node 
    // val (next1, m) = node1.next(level).get; var next = next1; var mark = m
    var next = readNode(node1.next(level).get)
    //myAssert(mark || next != null, node.toString)
    //while(mark){
    while(next == null){
      node1 = helpDelete(node1, level)
      // val (next1, m) = node1.next(level).get; next = next1; mark = m
      next = readNode(node1.next(level).get)
      // myAssert(mark || next != null, node1.toString)
    }
    // log("readNext ending: "+(n, next))
    (node1, next)
  }

  /** Scan for the next node at level with key at least key.  If necessary, help
    * with any deletions.
    * @return a pair (prev, next), such that prev.key < key <= next.key. */
  private def scanKey(node: SLNode, level: Int, key: Long): (SLNode, SLNode) = {
    // log("scanKey: "+node+" "+level+" "+key)
    myAssert(node != tail && node != null && node.key < key, node.toString)
    // Invariant: next = prev.next(level) (except for recent insertions),
    // prev.key < key
    val (n1, next1) = readNext(node, level); var prev = n1; var next = next1
    while(next.key < key){
      myAssert(next != tail, next.toString)
      prev = next
      val (p1, next1) = readNext(prev, level); prev = p1; next = next1
    }
    (prev, next) 
  }

  /** Pick a random level at which to insert a new node. */
  private def randomLevel: Int = {
    var l = 1
    while(scala.util.Random.nextBoolean && l < MaxLevel) l += 1
    l
  }

  private def backOff() = { } // FIXME

  /** Insert key0 into the queue. */
  def insert(key0: Int) = {
    // log("insert "+key0)
    val key = mkLongKey(key0)
    val level = randomLevel // newNode will be included in levels [0..level)
    val newNode = new SLNode(key, level)
    val savedNodes = new Array[SLNode](level) // nodes to insert after

    // Search for predecessors (I6-I9)
    var node = head
    for(l <- TopLevel to 1 by -1){
      node = scanKey(node, l, key)._1
      // val (node1, _) = scanKey(node, l, key); node = node1
      if(l < level) savedNodes(l) = node
    }

    // Try to connect at level 0 (I10-I30)
    var done = false
    do{
      val (prev, next) = scanKey(node, 0, key)
      // The paper assumes there are no repeated keys (I11-I24), but then
      // tests whether a key is repeated; we simply make this assumption.
      // myAssert(next.key != key, next.key+", "+key) // assumption of no repeated keys
      newNode.next(0).set(next, false)
      if(prev.next(0).compareAndSet((next, false), (newNode, false)))
        done = true
      else backOff()
    } while(!done)

    // Connect at intermediate levels (I31-I41)
    for(l <- 1 to level-1){
      // try to connext at level l
      newNode.validLevel = l; done = false // ; var count = 0
      var prev = savedNodes(l) // insert somewhere after prev
      do{
        // count += 1; myAssert(count < 100, prev.toString)
        val (prev1, next) = scanKey(prev, l, key); prev = prev1
        myAssert(prev.key <= key && key <= next.key)
        newNode.next(l).set(next, false)
        if(newNode.deleted.get ||
             prev.next(l).compareAndSet((next, false), (newNode, false)))
          done = true
        else backOff()
      } while(!done)
    }

    newNode.validLevel = level
    if(newNode.deleted.get) helpDelete(newNode, 0)
    // log("insert "+key0+" done\n"+this)
  }

  /** Return the minimum key. */
  def minimum : Option[Int] = {
    // log("minimum")
    var prev = head; var done = false; var node: SLNode = null
    // Search for first unmarked node, and try to mark it
    do{
      val (prev1, node1) = readNext(prev, 0); prev = prev1; node = node1
      if(node == tail){ /* log("minimum done: None\n"+this);  */ return None }
      myAssert(node != head)
      if(node == prev.next(0).get._1){
        if(!node.deleted.get){
          // myAssert(prev == head, prev.toString);
          done = true
        }
        else{
          myAssert(node.deleted.get, node.toString)
          helpDelete(node, 0)
          prev = head // correction to published version
        }
      } // end of if(node == prev.next(0).get)
    } while(!done);
    Some(extractIntKey(node.key))
  }

  /** Mark all next links from `node`, from `level` upwards.
    * Corresponds to D23-D26 (with level = 0), and H1-H4 */
  private def markNexts(node: SLNode, level: Int) = {
    var done = false // No need to reset done on each iteration
    for(l <- level until node.level){
      do{
        val (next, del) = node.next(l).get
        done = del || node.next(l).compareAndSet((next, false), (next, true))
      } while(!done);
    }
  }

  /** Remove node at level l; start gives the node to start searching from.
    * @return the predecessor of node, where searching can start at the next
    * level. */
  private def removeNode(node: SLNode, start: SLNode, l: Int) : SLNode = {
    // log("removeNode: start = "+start+"; l = "+l+"; next = "+node.next(l).get)
    myAssert(start != null, start.toString)
    var done = false; var prev = start // ; var count = 0
    do{
      // count += 1; myAssert(count < 100, "prev = "+prev+" node = "+node)
      if(node.next(l).get == (null, true)) done = true
      else{
        val (prev1, last) = scanKey(prev, l, node.key); prev = prev1
        val (next, del) = node.next(l).get
        if(last != node || (next == null && del)) done = true
        else{
          if(prev.next(l).compareAndSet((node, false), (next, false))){
            node.next(l).set(null, true); done = true
          }
          else if(node.next(l).get == (null, true)) done = true
          else backOff()
        }
      } // end of outer else
    } while(!done);
    prev
  }

  /** Delete and return the minimum key. */
  def deleteMin : Option[Int] = {
    // log("deleteMin")
    var prev = head; var done = false; var node: SLNode = null
    // Search for first unmarked node, and try to mark it
    do{
      val (prev1, node1) = readNext(prev, 0); prev = prev1; node = node1
      if(node == tail){ /* log("deleteMin done: None\n"+this); */ return None }
      // Try to mark node (D10-D22)
      myAssert(node != head)
      var retry = false // simulate "goto" in paper using loop
      do{
        retry = false; val del = node.deleted.get
        if(node == prev.next(0).get._1){
          if(!del){
            if(node.deleted.compareAndSet(false, true)){
              // myAssert(prev == head, prev.toString);
              node.prev = prev; done = true
            }
            else{ /* log("deleteMin: failed to mark "+node); */ retry = true }
          }
          else{
            myAssert(node.deleted.get)
            prev = helpDelete(node, 0)
            // log("deleteMin continuing from "+prev)
            prev = head // correction to published version
          }
        } // end of if(node == prev.next(0).get)
      } while(retry)
    } while(!done)

    // Mark all next links from node (D23-D26)
    markNexts(node, 0)

    // Complete deletion of node, by uncoupling (D27-D29)
    prev = head
    for(l <- node.level-1 to 0 by -1) prev = removeNode(node, prev, l)

    // Return result
    val key = extractIntKey(node.key) // ; log("deleteMin done: "+key+"\n"+this)
    Some(key)
  }

  /** Help with the deletion of node at level `level`. */
  private def helpDelete(node: SLNode, level: Int) : SLNode = {
    // log("helpDelete "+node);
    myAssert(node != head)
    // Mark all next links from node (H1-H4)
    markNexts(node, level)

    // Get node to start search (H5-H10)
    var prev = node.prev
    if(prev == null || level >= prev.validLevel){
      prev = head
      for(l <- MaxLevel-1 to level by -1){
        prev = scanKey(prev, l, node.key)._1
        // val (prev1, _) = scanKey(prev, l, node.key); prev = prev1
      }
    }

    // Remove node at this level, and return the previous node.
    prev = removeNode(node, prev, level)
    // log("helpDelete done "+prev)
    prev
  }

  def union(giver: binomialHeap.concurrent.Union) = ???
}
