package binomialHeap.concurrent

import ox.cads.util.{Log,NullLog}

/** Adding tidying to the binomial heap. */
abstract class Tidy(
  theLog: Log = NullLog, MaxRetries: Int = 2000,
  numMinsMin: Int = 1, numMinsDelMin: Int = 1,
  tidyRatio: Int = 8, RecTidy: Boolean = true,
  A: Int = 16, B: Int = 16,
  backtracking: Boolean = false, retrieving: Boolean = false,
  allowUnion: Boolean = false)
    extends
    DeleteMin(theLog, MaxRetries, numMinsMin, numMinsDelMin, tidyRatio, 
              RecTidy, A, B, backtracking, retrieving, allowUnion){

  /** Count of number of insert or successful deleteMin operations. */
  private val opCount = new java.util.concurrent.atomic.AtomicInteger(0)

  /** Maybe perform tidying. */
  protected def maybeTidy() =
    if(opCount.incrementAndGet%tidyRatio == 0) tidy()

  /* Our old strategy for tidying was as follows:  employ a thread tidying if
   * the number of roots reaches A; and employ an extra tidier for every
   * additional B roots. */

  // /** The number of threads tidying. */
  //private val tidyingCount = new AtomicInteger(0)

  //   val roots = rootCount.get; val tidiers = tidyingCount.get
  //   if(roots >= A+tidiers*B && tidyingCount.compareAndSet(tidiers, tidiers+1)){
  //     tidy(); tidyingCount.getAndDecrement
  //   }

  /** Maximum degree allowed for any node. */
  private val MaxDegree = 32

  // ========= Tidying

  /** Try to tidy, making use of merge and mergeSuccessiveTrees. */
  protected def tidy() = /* Profiler.time("tidy") */ {
    // Profiler.count("tidy")
    // Roots found so far, and their predecessor, indexed by degree:
    // buckets(d) holds a pair (pred,node) such that pred.next was node, and
    // node had degree d, or is null if no such pair found.
    val buckets = new Array[(Node,Node)](MaxDegree)

    // Try to merge curr, with predecessor pred, with a tree in buckets.
    // Return true if merge of curr succeeded.
    def tryMerge(pred: Node, curr: Node) : Boolean = {
      val currState = curr.maybeClearParent; val degree = currState.degree
      if(currState.next == null || (degree == 0 && random.nextInt(4) != 0)) false
        // don't merge last node or singletons: disrupts insert
      else{
        val predState = pred.maybeClearParent
        // Consider merging curr with a previous node
        if(buckets(degree) != null){
          val (predOther, other) = buckets(degree)
          val otherState = other.maybeClearParent
          // Test if states of other, predOther are suitable for a merger; if
          // not, maybe replace with (pred, curr)
          if(other == curr || !otherState.unlabelledParentless){
            if(predState.next == curr) buckets(degree) = (pred, curr)
            false
          }
          else if(otherState.degree != degree){// degree of other changed
            buckets(degree) = if(predState.next == curr) (pred, curr) else null
            if(RecTidy) tryMerge(predOther, other) // recurse
            false // merge of curr failed
          }
          else if(currState.unlabelledParentless){ // try merging
            val predOtherState = predOther.maybeClearParent
            if(other.key <= curr.key){
              if(predState.next == curr && predState.unlabelledParentless){
                // Try to make curr a child of other
                val ok =
                  merge(other, otherState, pred, predState, curr, currState)
                // if successful, try to merge recursively
                if(ok && RecTidy){
                  buckets(degree) = null
                  tryMerge(if(predOther == curr) pred else predOther, other)
                }
                ok
              } // end of if(predState.next == curr && predState.unlabelledParentless)
              else false // pred, curr no longer successive; forget this pair
            } // end of if(other.key <= curr.key)
            else if(predOtherState.next == other && predOtherState.unlabelledParentless){
              // Try to make other a child of curr
              val ok = merge(
                curr, currState, predOther, predOtherState, other, otherState)
              // If successful, try to merge recursively
              if(ok && RecTidy){
                buckets(degree) = null
                tryMerge(if(pred == other) predOther else pred, curr)
              }
              ok
            }
            else{ // predOther, other no longer successive
              if(predState.next == curr) buckets(degree) = (pred, curr)
              false
            }
          } // end of else if(currState.unlabelledParentless)
          else false // curr not unlabelled root; do nothing
        } // end of if(buckets(degree) != null)
        else{ // curr is first with this degree
          if(predState.next == curr) buckets(degree) = (pred, curr)
          false
        }
      }
    } // end of tryMerge

    // Traverse root list, trying to merge trees
    var curr = head
    while(curr != null){
      val (next, ok) = nextStep(curr)
      if(next != null && ok){
        val ok = tryMerge(curr, next)
        if(!ok || curr.getState.parent == next) curr = next
      }
      else curr = next
    }
  }

  /** Try to find the next node after curr to consider for merging.  The boolean
    * component of the result is true if the next node and curr are likely to
    * be successive roots, possibly after a current operation has finished. */
  private def nextStep(curr: Node) : (Node, Boolean) = {
    // Maybe skip to node after t2 if t2 is being merged.
    def skip(t1: Node, t2: Node, state2: NodeState) : (Node, Boolean) = {
      val myState2 = t2.getState
      if(myState2 == state2 || myState2.parent == t1) (state2.next, true)
      else (t2, true) // state of t2 has changed to interfere with merge
    }

    val currState = curr.maybeClearParent()
    if(currState.parent != null) (currState.parent, false)
    else currState.label match{
      case MergeNext(t1, t2, state2) => skip(t1, t2, state2)
      case MergeParent(pred, _, _, t2, state2) if curr == pred => 
         skip(curr, t2, state2)
      case Delete(_) => (currState.next, false)
      case _ => (currState.next, true)
    }
  }

  // ========= Merging

  /** Merge t1, with expected state state1, and t2, with expected state state2,
    * making t2 a child of t1.  pred is the expected predecessor of t2, with
    * expected state pState. 
    * @return true if successful. */
  private def merge(
    t1: Node, state1: NodeState, pred: Node, pState: NodeState,
    t2: Node, state2: NodeState) : Boolean
    = /* Profiler.notime("merge") */ {
    // Profiler.count("merge")
    myAssert(t1 != t2 && t1.key <= t2.key && state1.degree == state2.degree &&
               pState.next == t2 && state1.unlabelledParentless &&
               state2.unlabelledParentless && pState.unlabelledParentless)
    // Create labels
    val pLabel = if(t1 != pred) MergeNext(t1, t2, state2) else null
    val label = MergeParent(pred, pState, pLabel, t2, state2)
    // Try to add label to t1
    val state1L = state1.addLabel(label)
    if(t1.getState == state1 && pred.getState == pState &&
         t2.getState == state2 && t1.compareAndSet(state1, state1L)){
      if(t1 != pred)
        mergeLabelPred(t1, state1L, pred, pState, pLabel, t2, state2)
      else mergeUpdateT2(t1, state1L, pred, pState, t2, state2)
    }
    else false
  }

  /** Help with merger of non-successive trees by labelling pred.  t1 has been
    * labelled giving state1L.  pLabel needs to be added to pred.  Other
    * arguments as for merge. */
  protected def mergeLabelPred(
    t1: Node, state1L: NodeState, pred: Node, pState: NodeState,
    pLabel: MergeNext, t2: Node, state2: NodeState) : Boolean
    = {
    myAssert(t1.key <= t2.key && t1 != pred && state1L.degree == state2.degree &&
               pState.next == t2 && state1L.parent == null &&
               state1L.label.isInstanceOf[MergeParent] &&
               state2.unlabelledParentless && pState.unlabelledParentless)
    val pStateL = pState.addLabel(pLabel)
    if(pred.getState == pState && t2.getState == state2 &&
         pred.compareAndSet(pState, pStateL))
      mergeUpdateT2(t1, state1L, pred, pStateL, t2, state2) 
    else{
      if(t1.getState == state1L && (pred.getState.label ne pLabel) &&
           t2.getState.parent != t1)
        // Either pred labelling hasn't happened; or pred update has happened,
        // and t2 has been subsequently updated, meaning that the t1 update
        // has happened.  Put another way: the second conjunct ensures we're
        // not between the pred label and pred update steps; the thrid
        // conjunct ensures we're not between the t2 update and t1 update
        // steps.  Remove label from t1
        if(!t1.compareAndSet(state1L, state1L.addLabel(null)))
          // someone else removed label
          myAssert(! t1.getState.label.eq(state1L.label))
      false
    }
  }

  /** Help with merger of non-successive trees by performing t2 update.  t1 has
    * been labelled.  If t1 != pred then pred has been labelled; if t1 ==
    * pred, then the value of pStateL is irrelevant. */
  protected def mergeUpdateT2(
    t1: Node, state1L: NodeState, pred: Node, pStateL: NodeState,
    t2: Node, state2: NodeState)
      : Boolean = {
    myAssert(t1.key <= t2.key && state1L.degree == state2.degree &&
               state1L.parent == null && state2.unlabelledParentless &&
               state1L.label.isInstanceOf[MergeParent],
             "t1 = "+t1+"\npred = "+pred+"\n"+"pStateL = "+
               pStateL.toString(1,1,pred)+(t1==pred))
    myAssert(t1 == pred ||
               pStateL.next == t2 && pStateL.parent == null &&
               pStateL.label.isInstanceOf[MergeNext])
    val newNext = if(state1L.children.isEmpty) null else state1L.children.head
    val state2U = new NodeState(
      parent = t1, degree = state2.degree, 
      children = state2.children, next = newNext, seq = state2.seq+1)
    if(t2.getState == state2 && t2.compareAndSet(state2, state2U)){
      // rootCount.getAndDecrement
      if(t1 != pred)
        mergeUpdatePred(t1, state1L, pred, pStateL, t2, state2.next)
      else mergeUpdateT1Pred(t1, state1L, t2, state2.next)
      true // merge bound to succeed
    }
    else{
      if(t2.getState.parent != t1){
        // Either t2 has been updated in an interfering way, or the update on
        // t1 (and hence the whole operation) has completed, and some later
        // update has acted on t2.  Remove labels from pred and t1; the order
        // doesn't matter, here.
        if(t1 != pred){
          if(pred.getState == pStateL &&
               pred.compareAndSet(pStateL, pStateL.addLabel(null))){ }
          else myAssert(pred.getState.label ne pStateL.label)
        }
        if(t1.getState == state1L)
          t1.compareAndSet(state1L, state1L.addLabel(null))
        else myAssert(t1.getState.label ne state1L.label)
      }
      false
    } // end of outer else
  }

  /** Complete the merger, updating pred and then t1.  
    * Pre: the t2 update has happened.*/
  protected def mergeUpdatePred(
    t1: Node, state1L: NodeState, pred: Node, pStateL: NodeState,
    t2: Node, nextNode: Node) = {
    myAssert(t1.key <= t2.key && pStateL.next == t2 && state1L.parent == null &&
               state1L.label.isInstanceOf[MergeParent] &&
               pStateL.parent == null && pStateL.label.isInstanceOf[MergeNext])
    myAssert(t2.getState.parent == t1 ||
               t1.getState != state1L && pred.getState != pStateL)
    // update pred
    val pStateU = new NodeState(
      parent = null, degree = pStateL.degree, 
      children = pStateL.children, next = nextNode, seq = pStateL.seq)
    if(pred.getState == pStateL && pred.compareAndSet(pStateL, pStateU))
      mergeUpdateT1(t1, state1L, pred, t2)
  }

  /** Complete the merger, updating t1. */
  protected def mergeUpdateT1(t1: Node, state1L: NodeState, pred: Node, t2: Node)
    = {
    val state1U = new NodeState(
      degree = state1L.degree+1, children = t2 :: state1L.children, 
      next = state1L.next, seq = state1L.seq)
    if(t1.getState == state1L) t1.compareAndSet(state1L, state1U)
    // else another thread completed merger. 
  }

  /** Complete the merger of adjacent trees, updating t1. */
  protected def mergeUpdateT1Pred(
    t1: Node, state1L: NodeState, t2: Node, nextNode: Node)
    = {
    val state1U = new NodeState(
      degree = state1L.degree+1, children = t2 :: state1L.children,
      next = nextNode, seq = state1L.seq)
    if(t1.getState == state1L) t1.compareAndSet(state1L, state1U)
  }



}
