package binomialHeap.concurrent

import ox.cads.util.{Log,NullLog}

/** Adding the advance function, used in traversing, to the binomial heap. */
abstract class Traversing(
  theLog: Log = NullLog, MaxRetries: Int = 2000,
  numMinsMin: Int = 1, numMinsDelMin: Int = 1,
  tidyRatio: Int = 8, RecTidy: Boolean = true,
  A: Int = 16, B: Int = 16,
  backtracking: Boolean = false, retrieving: Boolean = false,
  allowUnion: Boolean = false)
    extends
    ConcHeapBase(theLog, MaxRetries, numMinsMin, numMinsDelMin, tidyRatio, 
                 RecTidy, A, B, backtracking, retrieving, allowUnion){

  // ========= Traversing

  /** Starting from n, get first node (possibly null) without Delete label, its
    * sequence number, and list of successive nodes with Delete label.  FIXME:
    * make tail recursive. */
  private def skipDeleted(n: Node) : (Node, Int, List[Node]) = {
    if(n == null) (n, -1, List())
    else{
      val nState = n.getState
      if(nState.label.isInstanceOf[Delete]){
        val (next, nextSeq, dels) = skipDeleted(nState.next)
        (next, nextSeq, n::dels)
      }
      else (n, nState.seq, List())
    }
  }

  /** Take the next step along the root list from a node that is a MergeNext or
    * MergeParent with t1=pred.
    * @param curr the current node
    * @param currState curr's (recent) state, with a MergeNext/MergeParent label.
    * @param t1 the t1 field from the label.
    * @param t2 the t2 field from the label. 
    * @param state2 the state2 field from the label. */
  private def skipMerge(t1: Node, t2: Node, state2: NodeState)
      : (Node, Int, List[Node]) = {
    myAssert(t1.key <= t2.key && state2.parent == null)
    val (next, nextSeq, dels) = skipDeleted(state2.next)
    val myState2 = t2.getState
    // The t1 update CAS hasn't happened yet.
    if(myState2 == state2) // t2 is unchanged; skip over t2
      (next, nextSeq, t2::dels)
    else if(myState2.parent == t1)
      // t2 CAS has happened.  Skip over t2 and advance to next.
      (next, nextSeq, dels)
    // else t2 has changed; advance as normal
    else if(myState2.label.isInstanceOf[Delete]){
      // t2 deleted; earlier call to skipDeleted has unreliable results
      val (next1, nextSeq1, dels1) = skipDeleted(myState2.next)
      (next1, nextSeq1, t2::dels1)
    }
    else (t2, myState2.seq, List()) 
  }    

  /** Try to find the next node to advance to from curr and its state; return
    * null if unsuccessful.  Also return any deleted node skipped over (or
    * null if none).  currSeq is expected to be curr.getSeq; if not, this will
    * fail.  In some circumstances, the operation will help another operation
    * (to ensure lock freedom), and then fail.
    * @return a triple (next, nextState, skipNode) where: next is the next node 
    * to advance to; nextState is the state read from next; skipNode is a node
    * skipped over, or null. */
  protected def advance(curr: Node, currSeq: Int) : (Node, Int, List[Node]) = {
    var result : (Node, Int, List[Node]) = null; var done = false
    var currState = curr.getState
    while(result == null && !done){
      if(currState.seq != currSeq) done = true// return null, restart traversal
      else{
        // currState.parent might have been decoupled from the list, but its
        // delete field not yet set.
        myAssert(currState.parent == null ||
                   currState.parent.getLabel.isInstanceOf[Delete],
                 "curr = "+curr.toString(2)+"\ncurrState = "+
                   currState.toString(2, 2, curr)+"\nparent = "+
                   currState.parent.toString(2))
        currState.label match{
          case null => result = skipDeleted(currState.next)
          case MergeParent(pred, _, _, t2, state2) =>
            if(pred == curr) result = skipMerge(curr, t2, state2)
            else result = skipDeleted(currState.next)
          case MergeNext(t1, t2, state2) => result = skipMerge(t1, t2, state2)
          case Delete(_) => assert(false)
          case _: HeapLabelList | _: UnionLast | _: Inserted =>
            result = skipDeleted(currState.next)
          case ug: UnionGiver =>
            result = skipDeleted(currState.next)
            if(ug.joined) result = (null, -1, List())
        } // end of match
        // check state not changed
        val newCurrState = curr.getState
        if(newCurrState != currState){ result = null; currState = newCurrState }
        // curr's state changed; try again
      } // end of else
    }
    // if(result != null && result._1 != null){
    //   val (next, nextSeq, _) = result
    //   val nextState = next.getState
    //   if(nextState.seq == nextSeq)
    //     myAssert(nextState.parent == null ||
    //              nextState.parent.getLabel.isInstanceOf[Delete],
    //              "currState = "+currState.toString(3, 3, curr, null)+
    //                "\nparent = "+currState.parent.toString(2))
    // }

    result
  }

}
