package binomialHeap.concurrent

import ox.cads.util.{Log,NullLog,ThreadID}

/** Adding the minimum operation to the binomial heap. */
abstract class Union(
  theLog: Log = NullLog, MaxRetries: Int = 2000,
  numMinsMin: Int = 1, numMinsDelMin: Int = 1,
  tidyRatio: Int = 8, RecTidy: Boolean = true,
  A: Int = 16, B: Int = 16,
  backtracking: Boolean = false, retrieving: Boolean = false,
  allowUnion: Boolean = false)
    extends
    Insert(theLog, MaxRetries, numMinsMin, numMinsDelMin, tidyRatio, 
         RecTidy, A, B, backtracking, retrieving, allowUnion){

  // ========= Union

  /** Add the elements of giver onto this, and clear giver.  If allowUnion is
    * false, the implementation assumes that there are no concurrent
    * operations on giver. */
  def union(giver: Union) = {
    assert(giver != this)
    if(allowUnion) concurrentUnion(giver) else sequentialUnion(giver)
  }

  // ========= Sequential union

  /** Add the elements of giver onto this, and clear giver, allowing concurrent
    * operations on this, but not on giver. */
  private def sequentialUnion(giver: Union) = {
    val giverFirst = giver.head.getState.next
    if(giverFirst != null){ // other heap not empty
      var done = false
      while(!done){
        val (last, lastState) = findLast
        myAssert(lastState.next == null && lastState.parentless &&
                   lastState.unlabelled)
        val newLastState = lastState.setNext(giverFirst)
        done = last.getState == lastState &&
          last.compareAndSet(lastState, newLastState)
      }
      giver.clear
      tidy()
    }
    // rootCount.getAndAdd(that.rootCount.get)
  }

  // ========= Concurrent union, operations on the giving heap.

  /** Prepare this heap to be the giving heap in a union.  If the heap is
    * non-empty, add a UnionGiver label to the head node to indicate that a
    * merge is happening.
    * @return the new status object and the first proper node (head.next). */
  private def labelGiverForUnion(receiver: Union): (UnionGiver, Node) = {
    // log("labelGiverForUnion "+(receiver.id, id))
    val ug = UnionGiver(receiver, new UnionReceiver(this))
    var first: Node = null; var done = false // first will be first node
    while(!done){
      val headState = head.getState
      if(headState.next == null) done = true // Empty heap
      else if(headState.label == null){
        if(head.compareAndSet(headState, headState.addLabel(ug))){
          first = headState.next; initUG(ug, first); done = true
        }
      }
      else{
        // log("labelGiverForUnion helping: "+headState.label);
        help(head, headState)
      }
    }
    (ug, first)
  }

  /** Initialise the UnionGiver label ug.  This must be called on the heap to
    * which ug applies. 
    * @param first the first proper node in the heap. */
  protected def initUG(ug: UnionGiver, first: Node): Unit = 
    ug.init(first, heapIndirection.get, epoch.get)

  /** Initialise the UnionGiver label ug.  This must be called on the heap to
    * which ug applies.  */
  protected def initUG(ug: UnionGiver): Unit = initUG(ug, head.getState.next)

  /** Complete the union on the giving heap.  Update the heapIndirection,
    * increment the epoch, and clear the label and next on head (as part of
    * completing the union on the giving heap). */
  private def completeUnionOnGiver(ug: UnionGiver) = {
    // Update headIndirection
    val hi = ug.heapIndirection; assert(hi != null)
    val newHi = ug.receiver.heapIndirection.get
    val headState = head.getState
    if(headState.label eq ug){
      // The receiver's status has not yet been updated as part of this union.
      if(heapIndirection.get == hi){
        // Note: only first thread to call the following two commands has an
        // effect.
        hi.setNext(newHi)
        heapIndirection.compareAndSet(
          hi, new HeapIndirection(this.asInstanceOf[ConcHeap]))
      }

      // Increment epoch
      val ep = ug.epoch; assert(ep >= 0)
      if(epoch.get == ep) epoch.compareAndSet(ep, ep+1)
      //   log(id+ ": incremented epoch of "+id+" from "+ep)
      // else
      //   log("failed to increment epoch of "+id+" from "+ep+"; found "+epoch.get)

      // Clear label and next on head
      if(head.getState == headState){
        assert(headState.next == ug.first)
        val newHeadState = new NodeState(
          degree = 0, children = List(), next = null,
          seq = headState.seq, label = null)
        head.compareAndSet(headState, newHeadState)
        // If the CAS fails, another thread has done an equivalent CAS
        assert(head.getLabel ne ug)
      }
      // else another thread has done an equivalent CAS
    } // end of if(head.getLabel eq ug)
    else{
      // log(id+": completeUnionOnGiver: status already cleared: "+head.getLabel)
      assert(epoch.get > ug.epoch,
             "this = "+this+"; epoch.get = "+epoch.get+
               "; ug.epoch = "+ug.epoch)
    }
  }

  // ========= Concurrent union, operations on the receiving heap.

  /** Add the elements of that onto this, and clear that, allowing concurrent
    * operations on both heaps. 
    * Start by setting status of giver. */
  private def concurrentUnion(that: Union): Unit = {
    // log("concurrentUnion:\n  receiver = "+this.toString+
    //      "\n  giver = "+that.toString)
    // Label that for the union; get first node and new status.
    val (ug, thatFirst) = that.labelGiverForUnion(this)
    if(thatFirst != null) unionLabelReceiver(that, thatFirst, ug, false)
    // else log("concurrentUnion done: empty")
  }

  /** Add UnionReceiver label to the head of the receiver.  This is done to
    * prevent concurrent unions affecting the same heap.  If there is a cycle
    * of conflicting unions, then roll one back.  If this union is rolled back
    * then re-start.
    * @param giver the giver heap.
    * @param giverFirst the first proper node of giver, non-null.
    * @param ug the status label on giver. */
  protected def unionLabelReceiver(
    giver: Union, giverFirst: Node, ug: UnionGiver, helping: Boolean)
      : Unit = {
    assert(((giver.head.getLabel eq ug) || ug.aborted || ug.joined) && ug.isInit)
    // log("unionLabelReceiver "+this.id)
    var done = false; val ur: UnionReceiver = ug.ur 
    while(!done){
      val headState = head.getState
      if(ug.aborted){
        assert(! ug.joined)
        // log("UnionGiver label rolled back; restarting")
        if(!helping) return concurrentUnion(giver) else return
      }
      // else if(ug.joined){
      //   log("UnionLabelReceiver: done by another thread, and LP reached"+ur.done)
      else if(ur.done){
        // log("UnionLabelReceiver: union finished ("+helping+")\n"+
        //       "Receiver = "+toString+"\n"+"Giver = "+giver.toString(10))
        return
      }
      //      }
      else headState.label match{
        case null  =>
          val hll = new HeapLabelList(this, ur)
          done = head.compareAndSet(headState, headState.addLabel(hll))
        /* If another thread aborts this union, it is from a state where head is
         * labelled with a UnionGiver label, and before head has ur added
         * to it.  Hence, if the above CAS, it cannot have been aborted after
         * ug.aborted is read.  A similar argument shows that the union
         * cannot have been completed and ur marked as done after the
         * call to ug.joined. */

        case hll: HeapLabelList =>
          // log("unionLabelreceiver adding to existing HeapLabelList "+hll)
          done = hll.add(ur)
          // log("unionLabelReceiver added to "+hll+"; "+done+"; "+ur.done)
        /* A similar argument to above shows that this union cannot have been 
         * aborted after ug.aborted is read.  A similar argument shows
         * that the union cannot have been completed and ur marked as
         * done before hll was added to head.  The call to addUR prevents
         * ur being added if it is already done. */

        case ug1: UnionGiver =>
          breakLoop
          if(!ug1.aborted && !ug.aborted){
            // log(ThreadID.get+"UnionLabelReceiver helping "+ug1)
            help(head, headState)
          }
          // else
            // log("UnionLabelReceiver: blocking UnionGiver removed or this aborted.")
            // Go back round the loop.  If ug1 has been aborted and replaced
            // by a UnionReceiver inside a HeapLabelList, we'll probably help
            // with that on the next round.  If ug has been aborted, we'll
            // deal with that as above.

        case _ => help(head, headState)
      } // end of match
    } // end of while(!done)

    unionJoin(giver, giverFirst, ug)
  }

  /** Try to detect a loop of UnionGiver statuses, rolling one back if so. */
  private def breakLoop = {
    // Follow the path of heaps linked by UnionGiver labels, looking for a
    // lasso.
    var heap = this
    // The heaps found so far, together with their UnionGiver labels. 
    var seen = List[(Union, UnionGiver)]()
    // done is true if non-loop found.  loop is true if loop found
    var done = false; var loop = false
    while(!done && !loop){
      // val headState = heap.head.getState
      heap.head.getLabel match{
        case ug @ UnionGiver(h, _) =>
          val pair = (heap, ug)
          if(seen.contains(pair)){
            // Loop found; set seen to be those pairs in the loop.
            loop = true; seen = pair :: seen.takeWhile(_ != pair)
          }
          else{ heap = h; seen ::= pair }
        case _ => done = true
      }
    }
    if(loop){
      // log("Loop of unions found: "+seen.reverse.map{case (h,ug) => (h.id, ug)})
      if(seen.forall{ case (h,ug) => h.head.getLabel eq ug }){
        // Pair with largest sequence number
        val (latest,latestUG) = seen.maxBy(_._2.seq)
        // The pair that is directly blocked by (latest,latestUG)
        val List((prev, prevUG)) = seen.filter(_._2.receiver == latest)
        // Make sure prevUG initialised
        if(!prevUG.isInit){
          // log("breakLoop helping to initialise "+prevUG)
          prev.initUG(prevUG)
        }
        // Mark latestUG as being aborted, to be replaced with prevUG.ur
        latestUG.abort(prevUG.ur)
        // Now replace the UnionGiver label with the blocked UnionReceiver label
        latest.replaceLabel(latestUG, prevUG.ur)
      }
      // else log("detected loop no longer exists")
    }
  }

  /** Replace label ug by ur.  This operation is performed on the giving heap.
    * Pre: the UnionGiver corresponding to ur has been initialised. */
  protected def replaceLabel(ug: UnionGiver, ur: UnionReceiver): Unit = {
    // log("replaceLabel: "+ug+"; "+ur)
    assert(!ug.joined)
    // Check the precondition.  The return below is an optimisation.
    ur.giver.head.getLabel match{
      case ug1 @ UnionGiver(_, ur1) if ur1 eq ur => assert(ug1.isInit)
      case _ =>
        // log("replaceLabel: corresponding UnionGiver removed")
        theLog.assert(head.getLabel ne ug); return
    }
    val hll = new HeapLabelList(this, ur)
    val headState = head.getState
    if((headState.label eq ug) &&
         head.compareAndSet(headState, headState.addLabel(hll))){ }
      // log("rolled back "+(id, ug)+"adding status "+ur)
    // Otherwise, another thread did the replacement
    assert(head.getLabel ne ug)
  }

  /** Locate the last node in this; update information in ug about it;
    * attempt to CAS last to join the heaps; update information in ug to
    * indicate this is done. 
    * @param giver the giving heap.
    * @param giverFirst the first node in giver. */
  protected def unionJoin(giver: Union, giverFirst: Node, ug: UnionGiver)
      : Unit = {
    // log("unionJoin"+(this.id, giver.id))
    var done = false
    while(!done){
      ug.getLastInfo match{
        case (null, null, false) =>
          // Find last node in this
          val (last, lastState) = findLast
          myAssert(lastState.next == null && lastState.parentless)
          val label = lastState.label
          myAssert(label == null ||
                     last == head && label.isInstanceOf[HeapLabelList])
          if(ug.setLastInfo(last, lastState))
            // Try to do the main CAS for the union
            done = updateLast(giverFirst, ug, last, lastState)

        case (last, lastState, d) => // Another thread found last.
          done = d || updateLast(giverFirst, ug, last, lastState)
      } // end of match
    } // end of while
    clearLast(giver, ug)
  }

  /** Try to update last, with state lastState, to point to giverFirst.  Also add
    * a UnionLast label to last, unless last == head, in which case retain the
    * label ur.  If successful, update the last field in ug to
    * indicate this has been done.  The last field in ug was previously
    * set to indicate last and lastState, and this will remain true unless and
    * until last's state changes to prevent the CAS from succeeding. */
  private def updateLast(giverFirst: Node, ug: UnionGiver,
      last: Node, lastState: NodeState) : Boolean = {
    // log("updateLast; last = "+last.toString)
    val lastLabel =
      if(head == last){
        val headLabel = lastState.label;
        assert(headLabel.isInstanceOf[HeapLabelList]); headLabel
      }
      else UnionLast(ug)
    val joinedLastState = new NodeState(
      degree = lastState.degree, children = lastState.children,
      next = giverFirst, seq = lastState.seq, label = lastLabel)
    // the following is the linearization point of the union if successful
    if(last.getState == lastState &&
         last.compareAndSet(lastState, joinedLastState)){
      // log("Joining CAS done")
      ug.setDone(last, joinedLastState); true
    }
    else{
      val newLastState = last.getState 
      // Note that if another thread did the CAS, we will have
      // newLastState.next == giverFirst until ug.joined is set.
      if(ug.joined){ /* log("updateLast CAS done by another thread"); */ true }
      else{
        // Interference, so retry in unionJoin.  The following assertion holds,
        // or else the cal to joined would have given true.
        assert(newLastState.next != giverFirst)
        // log("updateLast CAS interfered with by another thread: "+last.toString(2))
        ug.clearLastInfo(last, lastState); false
      }
    }
  } // end of updateLast

  /** Clear the label on the old last node of the receiver. 
    * @param giver the giving heap. */
  protected[concurrent] def clearLast(giver: Union, ug: UnionGiver) = {
    // log("clearLast")
    // Clear label on last
    val (last, lastState, true) = ug.getLastInfo
    if(last != head) clearLastLabel(last, lastState)
    completeUnion(giver, ug)
  }

  /** Clear the UnionLast label on last from lastState. */ 
  protected def clearLastLabel(last: Node, lastState: NodeState) = {
    //val (last, lastState, true) = ug.getLastInfo
    assert(lastState.label.isInstanceOf[UnionLast])
    if(last.getState == lastState)
      last.compareAndSet(lastState, lastState.addLabel(null))
    // If the CAS failed, a helping thread performed it.
  }

  /** Complete the union: clear the label and next reference of the giver's
    * head; clear the status of the giver.
    * @param giver the giving heap. */
  private def completeUnion(giver: Union, ug: UnionGiver) = {
    // log("completeUnion")
    // Update the heapIndirection, and clear label and next on giver's head.
    giver.completeUnionOnGiver(ug)
    val firstHeap = ug.first.getHeap
    // assert(false)
    myAssert(firstHeap == this || ug.ur.done,
             "this = "+id+"; giver = "+giver.id+"; firstHeap = "+firstHeap.id+
               "; ug = "+ug+"; first = "+ug.first.toString(3))
    
    // Clear status of this.
    clearUR(head, ug.ur)

    assert(!ug.aborted)
    // log("union done\n"+"Receiver = "+toString+"\n"+"Giver = "+giver.toString(10))
  }

  /** Clear the UnionReceiver label ur from node. */
  protected def clearUR(node: Node, ur: UnionReceiver) = {
    ur.done = true
    node.getLabel match{
      case hll: HeapLabelList => hll.remove(ur)
      case _ => {}
    }
  }
}
